"""
Utilities for fitting model PSFs to rate images.
"""

import logging
import os

import astropy.units as u
import numpy as np
import webbpsf
from astropy.modeling.fitting import LevMarLSQFitter
from astropy.nddata import CCDData, bitmask
from astropy.table import Table
from photutils.background import LocalBackground
from photutils.detection import DAOStarFinder
from photutils.psf import (
    GriddedPSFModel,
    IterativePSFPhotometry,
    PSFPhotometry,
    SourceGrouper,
)
from roman_datamodels.datamodels import ImageModel
from roman_datamodels.dqflags import pixel
from webbpsf import conf, gridded_library, restart_logging

__all__ = [
    "create_gridded_psf_model",
    "fit_psf_to_image_model",
    "dq_to_boolean_mask",
]

# set loggers to debug level by default:
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

# Phase C central wavelengths [micron], released by Goddard (Jan 2023):
# https://roman.ipac.caltech.edu/sims/Param_db.html#wfi_filters
filter_central_wavelengths = {
    "F062": 0.620,
    "F087": 0.869,
    "F106": 1.060,
    "F129": 1.293,
    "F146": 1.464,
    "F158": 1.577,
    "F184": 1.842,
    "F213": 2.125,
}

default_finder = DAOStarFinder(
    # these defaults extracted from the
    # romancal SourceDetectionStep
    fwhm=1.0,
    threshold=0.0,
    sharplo=0.0,
    sharphi=1.0,
    roundlo=-1.0,
    roundhi=1.0,
    peakmax=None,
)


def create_gridded_psf_model(
    path_prefix,
    filt,
    detector,
    oversample=11,
    fov_pixels=9,
    sqrt_n_psfs=2,
    overwrite=False,
    buffer_pixels=100,
    instrument_options=None,
    logging_level=None,
):
    """
    Compute a gridded PSF model for one SCA via
    `~webbpsf.gridded_library.CreatePSFLibrary`.

    Parameters
    ----------
    path_prefix : str or Path-like
        Prefix to the output file path for the gridded PSF model
        FITS file. A suffix denoting the detector name will
        be appended.
    filt : str
        Filter name, starting with "F". For example: `"F184"`.
    detector : str
        Computed gridded PSF model for this SCA.
        Examples include: `"SCA01"` or `"SCA18"`.
    oversample : int, optional
        Oversample factor, default is 11. See WebbPSF docs for details [1]_.
        Choosing an odd number makes the pixel convolution more accurate.
    fov_pixels : int, optional
        Field of view width [pixels]. Default is 12.
        See WebbPSF docs for details [1]_.
    sqrt_n_psfs : int, optional
        Square root of the number of PSFs to calculate, distributed uniformly
        across the detector. Default is 4.
    overwrite : bool, optional
        Overwrite output file if one already exists. Default is False.
    buffer_pixels : int, optional
        Calculate a grid of PSFs distributed uniformly across the detector
        at least ``buffer_pixels`` away from the detector edges. Default is 100.
    instrument_options : dict, optional
        Instrument configuration options passed to WebbPSF.
        For example, WebbPSF assumes Roman pointing jitter consistent with
        mission specs by default, but this can be turned off with:
        ``{'jitter': None, 'jitter_sigma': 0}``.
    logging_level : str, optional
        Set logging level by name if not `None`, otherwise inherit from
        the romancal logger.

    Returns
    -------
    gridmodel : `photutils.psf.GriddedPSFModel`
        Gridded PSF model evaluated at several locations on one SCA.
    model_psf_centroids : list of tuples
        Pixel locations of the PSF models calculated for ``gridmodel``.

    References
    ----------
    .. [1] `WebbPSF documentation for `webbpsf.JWInstrument.calc_psf`
       <https://webbpsf.readthedocs.io/en/latest/api/webbpsf.JWInstrument.html#webbpsf.JWInstrument.calc_psf>`_

    """
    if int(sqrt_n_psfs) != sqrt_n_psfs:
        raise ValueError(f"`sqrt_n_psfs` must be an integer, got {sqrt_n_psfs}.")
    n_psfs = int(sqrt_n_psfs) ** 2

    # webbpsf appends "_sca??.fits" to the requested path:
    expected_output_path = f"{path_prefix}_{detector.lower()}.fits"

    # Choose pixel boundaries for the grid of PSFs:
    start_pix = 0
    stop_pix = 4096

    # Choose locations on detector for each PSF:
    if sqrt_n_psfs != 1:
        pixel_range = np.linspace(
            start_pix + buffer_pixels, stop_pix - buffer_pixels, int(sqrt_n_psfs)
        )
    else:
        pixel_range = [(start_pix + stop_pix) / 2]

    # generate PSFs over a grid of detector positions [pix]
    model_psf_centroids = [(int(x), int(y)) for y in pixel_range for x in pixel_range]

    if not os.path.exists(expected_output_path) or overwrite:
        if logging_level is None:
            # pass along logging level from __name__'s logger to WebbPSF:
            logging_level = logging.getLevelName(log.level)

        # set the WebbPSF logging level (similar to webbpsf.utils.setup_logging):
        conf.logging_level = logging_level
        restart_logging(verbose=False)

        wfi = webbpsf.roman.WFI()
        wfi.filter = filt

        if instrument_options is not None:
            wfi.options.update(instrument_options)

        # Initialize the PSF library
        inst = gridded_library.CreatePSFLibrary(
            instrument=wfi,
            filter_name=filt,
            detectors=detector.upper(),
            num_psfs=n_psfs,
            oversample=oversample,
            fov_pixels=fov_pixels,
            add_distortion=False,
            crop_psf=False,
            save=True,
            filename=path_prefix,
            overwrite=overwrite,
            verbose=False,
        )

        inst.location_list = model_psf_centroids

        # Create the PSF grid:
        gridmodel = inst.create_grid()

    elif os.path.exists(expected_output_path):
        logging.log(
            logging.INFO,
            f"Loading existing gridded PSF model from {expected_output_path}",
        )
        psf_model = CCDData.read(expected_output_path, unit=u.electron / u.s, ext=0)
        psf_model.meta = dict(psf_model.meta)
        psf_model.meta["oversampling"] = oversample
        psf_model.meta["grid_xypos"] = np.array(
            [list(tup)[::-1] for tup in model_psf_centroids]
        )
        gridmodel = GriddedPSFModel(psf_model)

    return gridmodel, model_psf_centroids


def fit_psf_to_image_model(
    image_model=None,
    data=None,
    error=None,
    dq=None,
    photometry_cls=PSFPhotometry,
    psf_model=None,
    grouper=None,
    fitter=None,
    localbkg_estimator=None,
    finder=None,
    x_init=None,
    y_init=None,
    progress_bar=False,
    error_lower_limit=None,
    fit_shape=(15, 15),
    exclude_out_of_bounds=True,
):
    """
    Fit PSF models to an ``ImageModel``.

    Parameters
    ----------
    image_model : `roman_datamodels.datamodels.ImageModel`
        Image datamodel. If ``image_model`` is supplied,
        ``data,error,dq`` should be `None`.
    data : `astropy.units.Quantity`
        Fit a PSF model to the rate image ``data``.
        If ``data,error,dq`` are supplied, ``image_model`` should be `None`.
    error : `astropy.units.Quantity`
        Uncertainties on fluxes in ``data``. Should be `None` if
        ``image_model`` is supplied.
    dq : `numpy.ndarray`
        Data quality bitmask for ``data``. Should be `None` if
        ``image_model`` is supplied.
    photometry_cls : {`photutils.psf.PSFPhotometry`,
            `photutils.psf.IterativePSFPhotometry`}
        Choose a photutils PSF photometry technique (default or iterative).
    psf_model : `astropy.modeling.Fittable2DModel`
        The 2D PSF model to fit to the rate image. Usually this model is an instance
        of `photutils.psf.GriddedPSFModel`.
    grouper : `photutils.psf.SourceGrouper`
        Specifies rules for attempting joint fits of multiple PSFs when
         there are nearby sources at small separations.
    fitter : `astropy.modeling.fitting.Fitter`, optional
        Modeling class which optimizes the PSF fit.
        Default is `astropy.modeling.fitting.LevMarLSQFitter(calc_uncertainties=True)`.
    localbkg_estimator : `photutils.background.LocalBackground`, optional
        Specifies inner and outer radii for computing flux background near
        a source. Default has ``inner_radius=10, outer_radius=30``.
    finder : subclass of `photutils.detection.StarFinderBase`, optional
        When ``photutils_cls`` is `photutils.psf.IterativePSFPhotometry`, the
        ``finder`` is called to determine if sources remain in the rate image
        after one PSF model is fit to the observations and removed.
        Default was extracted from the `DAOStarFinder` call in the
        Source Detection step.
    x_init : `numpy.ndarray`, optional
        Initial guesses for the ``x`` pixel coordinates of each source to fit.
    y_init : `numpy.ndarray`, optional
        Initial guesses for the ``y`` pixel coordinates of each source to fit.
    progress_bar : bool, optional
        Render a progress bar via photutils. Default is False.
    error_lower_limit : `astropy.units.Quantity`, optional
        Since some synthetic images may have bright sources with very
        small statistical uncertainties, the ``error`` can be clipped at
        ``error_lower_limit`` to prevent over-confident fits.
    fit_shape : int, or tuple of length 2, optional
        Rectangular shape around the center of a star that will
        be used to define the PSF-fitting data. See docs for
        `photutils.psf.PSFPhotometry` for details. Default is ``(16, 16)``.
    exclude_out_of_bounds : bool, optional
        If `True`, do not attempt to fit stars which have initial centroids
        that fall outside the pixel limits of the SCA. Default is False.

    Returns
    -------
    results_table : `astropy.table.QTable`
        PSF photometry results.
    photometry : instance of class ``photutils_cls``
        PSF photometry instance with configuration settings and results.

    """
    if grouper is None:
        # minimum separation before sources are fit simultaneously:
        grouper = SourceGrouper(min_separation=5)  # [pix]

    if fitter is None:
        fitter = LevMarLSQFitter(calc_uncertainties=True)

    # the iterative PSF method requires a finder:
    psf_photometry_kwargs = {}
    if photometry_cls is IterativePSFPhotometry or (x_init is None and y_init is None):
        if finder is None:
            finder = default_finder
        psf_photometry_kwargs["finder"] = finder

    if localbkg_estimator is None:
        localbkg_estimator = LocalBackground(
            inner_radius=10,  # [pix]
            outer_radius=30,  # [pix]
        )

    photometry = photometry_cls(
        grouper=grouper,
        localbkg_estimator=localbkg_estimator,
        psf_model=psf_model,
        fitter=fitter,
        fit_shape=fit_shape,
        aperture_radius=fit_shape[0],
        progress_bar=progress_bar,
        **psf_photometry_kwargs,
    )

    if x_init is not None and y_init is not None:
        guesses = Table(np.column_stack([x_init, y_init]), names=["x_init", "y_init"])
    else:
        guesses = None

    if image_model is None:
        if data is None and error is None:
            raise ValueError(
                "PSF fitting requires either an ImageModel, "
                "or arrays for the data and error."
            )

    ignore_flags = pixel.NO_LIN_CORR
    # presently the linearity correction is somewhat problematic in
    # CRDS reference files; we should replace this with ignore_flags = 0
    # at some point in the future.

    if dq is None:
        if image_model is not None:
            mask = dq_to_boolean_mask(image_model, ignore_flags=ignore_flags)
        else:
            mask = None
    else:
        mask = dq_to_boolean_mask(dq)

    if data is None and image_model is not None:
        data = image_model.data

    if error is None and image_model is not None:
        error = image_model.err

    if error_lower_limit is not None:
        # option to enforce a lower limit on the flux uncertainties
        error = np.clip(error, error_lower_limit, None)

    # we also mask non-finite values in the data and error arrays:
    non_finite = ~np.isfinite(data) | ~np.isfinite(error)

    if exclude_out_of_bounds and guesses is not None:
        # don't attempt to fit PSFs for objects with initial centroids
        # outside the detector boundaries:
        init_centroid_in_range = (
            (guesses["x_init"] > 0)
            & (guesses["x_init"] < data.shape[0])
            & (guesses["y_init"] > 0)
            & (guesses["y_init"] < data.shape[1])
        )
        guesses = guesses[init_centroid_in_range]

    # fit the model PSF to the data:
    results_table = photometry(
        data=data, error=error, init_params=guesses, mask=mask | non_finite
    )

    # results are stored on the PSFPhotometry instance:
    return results_table, photometry


def dq_to_boolean_mask(image_model_or_dq, ignore_flags=0, flag_map_name="ROMAN_DQ"):
    """
    Convert a DQ bitmask to a boolean mask. Useful for photutils methods.

    Parameters
    ----------
    image_model_or_dq : `roman_datamodels.datamodels.ImageModel` or `numpy.ndarray`
        ImageModel containing the DQ bitmask to convert to a boolean mask,
        or the DQ bitmask itself.
    ignore_flags : int, str, list, None (default = 0)
        See docs for `astropy.nddata.bitmask.extend_bit_flag_map`
    flag_map_name : str
        Name for the bitmask flag map in the astropy bitmask registry

    Returns
    -------
    mask : `numpy.ndarray`
        Boolean mask
    """

    if isinstance(image_model_or_dq, ImageModel):
        dq = image_model_or_dq.dq
    else:
        dq = image_model_or_dq

    # add the Roman DQ flags to the astropy bitmask registry:
    dq_flag_map = {dq.name: dq.value for dq in pixel if dq.name != "GOOD"}
    bitmask.extend_bit_flag_map(flag_map_name, **dq_flag_map)

    # convert the bitmask to a boolean mask:
    mask = bitmask.bitfield_to_boolean_mask(dq, ignore_flags=ignore_flags)
    return mask.astype(bool)



import numpy as np
import pytest
from metrics_logger.decorators import metrics_logger
from roman_datamodels import datamodels as rdm

from romancal.resample.resample_step import ResampleStep
from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@metrics_logger("DMS342", "DMS343", "DMS344", "DMS345")
@pytest.mark.bigdata
def test_resample_single_file(rtdata, ignore_asdf_paths):
    input_data = [
        "r0000101001001001001_01101_0001_WFI01_cal.asdf",
        "r0000101001001001001_01101_0002_WFI01_cal.asdf",
    ]
    output_data = "mosaic_resamplestep.asdf"

    [rtdata.get_data(f"WFI/image/{data}") for data in input_data]
    asnfn = "mosaic_asn.json"
    rtdata.get_data(f"WFI/image/{asnfn}")
    rtdata.get_truth(f"truth/WFI/image/{output_data}")

    rtdata.input = asnfn
    rtdata.output = output_data

    # instantiate ResampleStep (for running and log access)
    step = ResampleStep()

    args = [
        "romancal.step.ResampleStep",
        rtdata.input,
        "--rotation=0",
        f"--output_file='{rtdata.output}'",
    ]
    RomanStep.from_cmdline(args)
    resample_out = rdm.open(rtdata.output)

    step.log.info(
        "ResampleStep recorded as complete? :"
        f' {resample_out.meta.cal_step.resample == "COMPLETE"}'
    )
    assert resample_out.meta.cal_step.resample == "COMPLETE"

    step.log.info(
        "ResampleStep created 'meta.resample'? :"
        f' {hasattr(resample_out.meta, "resample")}'
    )
    assert hasattr(resample_out.meta, "resample")

    step.log.info(
        f"""DMS342 MSG: Was ICRS used as the mosaic astrometric reference frame? :\
            {
                resample_out.meta.coordinates.reference_frame == "ICRS"
            }
        """
    )
    assert resample_out.meta.coordinates.reference_frame == "ICRS"

    step.log.info(
        f"""DMS343 MSG: ResampleStep created new attribute data quality information? :\
            {
                all(
                    hasattr(resample_out, x) for x in [
                        "data",
                        "err",
                        "var_poisson",
                        "var_rnoise",
                        "var_flat",
                    ]
                )
            }"""
    )
    assert all(
        hasattr(resample_out, x)
        for x in ["data", "err", "var_poisson", "var_rnoise", "var_flat"]
    )

    step.log.info(
        f"""DMS343 MSG: Were the variance arrays populated (variance propagation)? :\
            {
                all(
                    np.sum(~np.isnan(getattr(resample_out, x))) for x in [
                        "var_poisson",
                        "var_rnoise",
                    ]
                )
            }"""  # noqa: E501
    )
    assert all(
        np.sum(~np.isnan(getattr(resample_out, x)))
        for x in ["var_poisson", "var_rnoise"]
    )

    step.log.info(
        f"""DMS343 MSG: Are there NaNs or zeros in the variance arrays, indicating poor data quality? :\
            {
                any(
                    np.sum(
                        np.logical_or(
                            np.isnan(getattr(resample_out, x)),
                            np.equal(getattr(resample_out, x), 0)
                        )
                    ) > 0 for x in ["var_poisson", "var_rnoise", "var_flat"]
                )

            }"""
    )
    assert all(
        np.sum(np.isnan(getattr(resample_out, x)))
        for x in ["var_poisson", "var_rnoise", "var_flat"]
    )

    step.log.info(
        f"""DMS344 MSG: ResampleStep created new attribute with total exposure time? :\
            {"product_exposure_time" in resample_out.meta.resample}
        """
    )
    assert "product_exposure_time" in resample_out.meta.resample

    step.log.info(
        f"""DMS345 MSG: ResampleStep included all metadata relevant to the creation of the mosaic? :\
            {
                all(
                    hasattr(resample_out.meta.resample, x)
                    and bool(getattr(resample_out.meta.resample, x))
                    for x in [
                        "pixel_scale_ratio",
                        "pixfrac",
                        "pointings",
                        "product_exposure_time",
                        "weight_type",
                        "members",
                    ]
                )
            }"""  # noqa: E501
    )
    assert all(
        hasattr(resample_out.meta.resample, x)
        and bool(getattr(resample_out.meta.resample, x))
        for x in [
            "pixel_scale_ratio",
            "pixfrac",
            "pointings",
            "product_exposure_time",
            "weight_type",
            "members",
        ]
    )

    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    step.log.info("Was the proper Resample data produced?" f" : {diff.identical}")
    assert diff.identical, diff.report()



""" Tests for the saturation step"""

import os

import pytest
import roman_datamodels as rdm

from romancal.step import SaturationStep
from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@pytest.mark.bigdata
def test_saturation_image_step(rtdata, ignore_asdf_paths):
    """Testing retrieval of best ref file for image data,
    and creation of a ramp file with CRDS selected saturation file applied."""

    input_file = "r0000101001001001001_01101_0001_WFI01_dqinit.asdf"
    rtdata.get_data(f"WFI/image/{input_file}")
    rtdata.input = input_file

    # Test CRDS
    step = SaturationStep()
    model = rdm.open(rtdata.input)
    ref_file_path = step.get_reference_file(model, "saturation")
    ref_file_name = os.path.split(ref_file_path)[-1]

    # Confirm that a WFI saturation reference file name was returned
    assert "roman_wfi_saturation" in ref_file_name

    # Test SaturationStep
    output = "r0000101001001001001_01101_0001_WFI01_saturation.asdf"
    rtdata.output = output

    args = ["romancal.step.SaturationStep", rtdata.input]
    RomanStep.from_cmdline(args)

    ramp_out = rdm.open(rtdata.output)
    assert "roman.pixeldq" in ramp_out.to_flat_dict()

    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()


@pytest.mark.bigdata
def test_saturation_grism_step(rtdata, ignore_asdf_paths):
    """Testing retrieval of best ref file for grism data,
    and creation of a ramp file with CRDS selected saturation file applied."""

    input_file = "r0000201001001001001_01101_0001_WFI01_dqinit.asdf"
    rtdata.get_data(f"WFI/grism/{input_file}")
    rtdata.input = input_file

    # Test CRDS
    step = SaturationStep()
    model = rdm.open(rtdata.input)
    ref_file_path = step.get_reference_file(model, "saturation")
    ref_file_name = os.path.split(ref_file_path)[-1]

    # Confirm that a WFI saturation reference file name was returned
    assert "roman_wfi_saturation" in ref_file_name

    # Test SaturationStep
    output = "r0000201001001001001_01101_0001_WFI01_saturation.asdf"
    rtdata.output = output

    args = ["romancal.step.SaturationStep", rtdata.input]
    RomanStep.from_cmdline(args)

    ramp_out = rdm.open(rtdata.output)
    assert "roman.pixeldq" in ramp_out.to_flat_dict()

    rtdata.get_truth(f"truth/WFI/grism/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()



""" Roman tests for flat field correction """

import copy

import numpy as np
import pytest
import roman_datamodels as rdm
from gwcs.wcstools import grid_from_bounding_box
from metrics_logger.decorators import metrics_logger
from numpy.testing import assert_allclose
from roman_datamodels.dqflags import pixel

from romancal.assign_wcs.assign_wcs_step import AssignWcsStep
from romancal.pipeline.exposure_pipeline import ExposurePipeline

from .regtestdata import compare_asdf


def passfail(bool_expr):
    """set pass fail"""
    if bool_expr:
        return "Pass"
    return "Fail"


@pytest.mark.bigdata
@pytest.mark.soctests
@metrics_logger("DMS86", "DMS87", "DMS88", "DMS361", "DMS280")
def test_level2_image_processing_pipeline(rtdata, ignore_asdf_paths):
    """Tests for flat field imaging processing requirements DMS86 & DMS 87"""
    input_data = "r0000101001001001001_01101_0001_WFI01_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "r0000101001001001001_01101_0001_WFI01_cal.asdf"
    rtdata.output = output
    args = [
        "--disable-crds-steppars",
        "roman_elp",
        rtdata.input,
    ]
    ExposurePipeline.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()

    # Perform DMS tests
    # Initial prep
    model = rdm.open(rtdata.output, lazy_load=False)
    pipeline = ExposurePipeline()

    # DMS280 result is an ImageModel
    pipeline.log.info(
        "DMS280 MSG: Testing that result is a Level 2 model......."
        + passfail(isinstance(model, rdm.datamodels.ImageModel))
    )

    # DMS86 instrument artifact correction tests
    pipeline.log.info(
        "Status of the step:             assign_wcs    "
        + str(model.meta.cal_step.assign_wcs)
    )
    pipeline.log.info(
        "DMS86 MSG: Testing completion of wcs assignment inLevel 2 image output......."
        + passfail(model.meta.cal_step.assign_wcs == "COMPLETE")
    )
    assert model.meta.cal_step.assign_wcs == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             flat_field    "
        + str(model.meta.cal_step.flat_field)
    )
    pipeline.log.info(
        "DMS86 MSG: Testing completion of flat fielding inLevel 2 image output......."
        + passfail(model.meta.cal_step.flat_field == "PASS")
    )
    assert model.meta.cal_step.flat_field == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             dark          " + str(model.meta.cal_step.dark)
    )
    pipeline.log.info(
        "DMS86 MSG: Testing completion of dark correction inLevel 2 image output......."
        + passfail(model.meta.cal_step.dark == "COMPLETE")
    )
    assert model.meta.cal_step.dark == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             dq_init       "
        + str(model.meta.cal_step.dq_init)
    )
    pipeline.log.info(
        "DMS86 MSG: Testing completion of data quality correction in Level 2 image"
        " output......." + passfail(model.meta.cal_step.dq_init == "COMPLETE")
    )
    assert model.meta.cal_step.dq_init == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             jump          " + str(model.meta.cal_step.jump)
    )
    pipeline.log.info(
        "DMS86 MSG: Testing completion of jump detection inLevel 2 image output......."
        + passfail(model.meta.cal_step.jump == "COMPLETE")
    )
    assert model.meta.cal_step.jump == "COMPLETE"
    uneven = len({len(x) for x in model.meta.exposure.read_pattern}) > 1
    pipeline.log.info(
        "DMS361: Testing that jump detection detected jumps in uneven ramp in "
        "Level 2 image output......."
        + passfail(uneven & np.any(model.dq & pixel.JUMP_DET))
    )
    assert uneven & np.any(model.dq & pixel.JUMP_DET)
    pipeline.log.info(
        "Status of the step:             linearity     "
        + str(model.meta.cal_step.linearity)
    )
    pipeline.log.info(
        "DMS86 MSG: Testing completion of linearity correction in Level 2 image"
        " output......." + passfail(model.meta.cal_step.linearity == "COMPLETE")
    )
    assert model.meta.cal_step.linearity == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             ramp_fit      "
        + str(model.meta.cal_step.ramp_fit)
    )
    pipeline.log.info(
        "DMS86 MSG: Testing completion of ramp fitting inLevel 2 image output......."
        + passfail(model.meta.cal_step.ramp_fit == "COMPLETE")
    )
    assert model.meta.cal_step.ramp_fit == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             saturation    "
        + str(model.meta.cal_step.saturation)
    )
    pipeline.log.info(
        "DMS86 MSG: Testing completion of saturation detection in Level 2 image"
        " output......." + passfail(model.meta.cal_step.saturation == "COMPLETE")
    )
    assert model.meta.cal_step.saturation == "COMPLETE"

    # DMS-129 tests
    # check if assign_wcs step is complete
    pipeline.log.info(
        "DMS-129 MSG: Status of the step:             assign_wcs    "
        + str(model.meta.cal_step.assign_wcs)
    )
    pipeline.log.info(
        "DMS-129 MSG: Testing completion of WCS assignment inLevel 2 image"
        " output......." + passfail(model.meta.cal_step.assign_wcs == "COMPLETE")
    )
    assert model.meta.cal_step.assign_wcs == "COMPLETE"
    # check if WCS exists
    pipeline.log.info("DMS-129 MSG: Testing that a WCS object exists    ")
    pipeline.log.info(
        "DMS-129 MSG: Testing that WCS exists inLevel 2 image output......."
        + passfail(model.meta.wcs is not None)
    )
    assert model.meta.wcs is not None
    pipeline.log.info(
        "DMS-129 MSG: Testing that geometric distortion information is available"
        " inLevel 2 image output......."
        + passfail("v2v3" in model.meta.wcs.available_frames)
    )
    assert "v2v3" in model.meta.wcs.available_frames
    # compare coordinates before and after distortion correction has been applied
    # 1 - get new image array based on the model
    x0, y0 = grid_from_bounding_box(model.meta.wcs.bounding_box)
    # 2 - apply the distortion-corrected WCS solution to new image array
    corrected_coords = model.meta.wcs(x0, y0)
    # 3 - apply the transformation from 'v2v3' to 'world' without distortion correction
    original_coords = model.meta.wcs.get_transform("v2v3", "world")(x0, y0)
    # compare both results to make sure they don't match
    # (which means the distortion correction was actually applied to the model)
    pipeline.log.info(
        "DMS-129 MSG: Testing that distortion correction was applied toLevel 2 image"
        " output......."
        + passfail(
            (corrected_coords[0] != original_coords[0]).all()
            & (corrected_coords[1] != original_coords[1]).all()
        )
    )
    assert (corrected_coords[0] != original_coords[0]).all()
    assert (corrected_coords[1] != original_coords[1]).all()

    # DMS87 data quality tests
    pipeline.log.info(
        "DMS87 MSG: Testing existence of data quality array (dq) in Level 2 image"
        " output......." + passfail("dq" in model.keys())
    )
    assert "dq" in model.keys()
    pipeline.log.info(
        "DMS87 MSG: Testing existence of general error array (err) in Level 2 image"
        " output......." + passfail("err" in model.keys())
    )
    assert "err" in model.keys()
    pipeline.log.info(
        "DMS87 MSG: Testing existence of Poisson noise variancearray (var_poisson) in"
        " Level 2 image output......." + passfail("var_poisson" in model.keys())
    )
    assert "var_poisson" in model.keys()
    pipeline.log.info(
        "DMS87 MSG: Testing existence of read noise variance array (var_rnoise) in"
        " level 2 image output......." + passfail("var_rnoise" in model.keys())
    )
    assert "var_rnoise" in model.keys()
    pipeline.log.info(
        "DMS87 MSG: Testing existence of flatfield uncertainty variance array"
        " (var_flat) in Level 2 image output...." + passfail("var_flat" in model.keys())
    )
    assert "var_flat" in model.keys()

    # DMS88 total exposure time test
    pipeline.log.info(
        "DMS88 MSG: Testing existence of total exposure time (exposure_time) in Level 2"
        " image output......." + passfail("exposure_time" in model.meta.exposure)
    )
    assert "exposure_time" in model.meta.exposure

    # DMS-136 PSF tests
    pipeline.log.info(
        "DMS-136 MSG: Testing existence of  detector and "
        "optical element (detector & optical_element) in Level 2 "
        "image output......." + passfail("exposure_time" in model.meta.exposure)
    )
    assert "detector" in model.meta.instrument
    assert "optical_element" in model.meta.instrument

    # DMS89 WCS tests
    pipeline.log.info(
        "DMS89 MSG: Testing that the wcs boundingbox was generated......."
        + passfail(len(model.meta.wcs.bounding_box) == 2)
    )
    assert len(model.meta.wcs.bounding_box) == 2

    # Save original wcs information
    orig_wcs = copy.deepcopy(model.meta.wcs)
    del model.meta["wcs"]

    # Create new pointing for the model
    # RA & Dec are each shifted + 10 degrees, unless they are near
    # the upper limit, in which case they are shifted -10 degrees.
    delta = [10.0, 10.0]
    if model.meta.wcsinfo.ra_ref >= 350.0:
        delta[0] *= -1.0

    if model.meta.wcsinfo.dec_ref >= 80.0:
        delta[1] *= -1.0

    model.meta.wcsinfo.ra_ref += delta[0]
    model.meta.wcsinfo.dec_ref += delta[1]

    # Create new wcs object for the new pointing
    model = AssignWcsStep.call(model)

    rtdata.output = output.rsplit(".", 1)[0] + "_repoint.asdf"
    model.to_asdf(rtdata.output)

    # Test that repointed file matches truth
    rtdata.get_truth("truth/WFI/image/" + output.rsplit(".", 1)[0] + "_repoint.asdf")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()

    pipeline.log.info(
        "DMS89 MSG: Testing that the different pointings create differing wcs......."
        + passfail(
            ((np.abs(orig_wcs(2048, 2048)[0] - model.meta.wcs(2048, 2048)[0])) - 10.0)
            < 1.0
        )
    )
    assert_allclose(
        [orig_wcs(2048, 2048)[0] + delta[0], orig_wcs(2048, 2048)[1] + delta[1]],
        model.meta.wcs(2048, 2048),
        atol=1.0,
    )


@pytest.mark.bigdata
@pytest.mark.soctests
@metrics_logger("DMS278", "DMS90", "DMS91", "DMS9", "DMS365")
def test_level2_grism_processing_pipeline(rtdata, ignore_asdf_paths):
    """Tests for flat field grism processing requirements DMS90, DMS91 and DMS 278"""
    input_data = "r0000201001001001001_01101_0001_WFI01_uncal.asdf"
    rtdata.get_data(f"WFI/grism/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "r0000201001001001001_01101_0001_WFI01_cal.asdf"
    rtdata.output = output
    args = [
        "--disable-crds-steppars",
        "roman_elp",
        rtdata.input,
    ]
    ExposurePipeline.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/grism/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()

    # Perform DMS tests
    # Initial prep
    model = rdm.open(rtdata.output, lazy_load=False)
    pipeline = ExposurePipeline()

    # DMS90 instrument artifact correction tests
    pipeline.log.info(
        "Status of the step:             assign_wcs    "
        + str(model.meta.cal_step.assign_wcs)
    )
    pipeline.log.info(
        "DMS90 MSG: Testing completion of wcs assignment in Level 2 spectroscopic"
        " output......." + passfail(model.meta.cal_step.assign_wcs == "COMPLETE")
    )
    assert model.meta.cal_step.assign_wcs == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             flat_field    "
        + str(model.meta.cal_step.flat_field)
    )
    pipeline.log.info(
        "DMS90 MSG: Testing expected skip of flat fielding in Level 2 spectroscopic"
        " output......." + passfail(model.meta.cal_step.flat_field == "SKIPPED")
    )
    assert model.meta.cal_step.flat_field == "SKIPPED"
    pipeline.log.info(
        "Status of the step:             dark          " + str(model.meta.cal_step.dark)
    )
    pipeline.log.info(
        "DMS90 MSG: Testing completion of dark correction in Level 2 spectroscopic"
        " output......." + passfail(model.meta.cal_step.dark == "COMPLETE")
    )
    assert model.meta.cal_step.dark == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             dq_init       "
        + str(model.meta.cal_step.dq_init)
    )
    pipeline.log.info(
        "DMS90 MSG: Testing completion of data quality correction in Level 2"
        " spectroscopic output......."
        + passfail(model.meta.cal_step.dq_init == "COMPLETE")
    )
    assert model.meta.cal_step.dq_init == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             jump          " + str(model.meta.cal_step.jump)
    )
    pipeline.log.info(
        "DMS90 MSG: Testing completion of jump detection in Level 2 spectroscopic"
        " output......." + passfail(model.meta.cal_step.jump == "COMPLETE")
    )
    assert model.meta.cal_step.jump == "COMPLETE"
    uneven = len({len(x) for x in model.meta.exposure.read_pattern}) > 1
    pipeline.log.info(
        "DMS365: Testing that jump detection detected jumps in uneven ramp in "
        "Level 2 image output......."
        + passfail(uneven & np.any(model.dq & pixel.JUMP_DET))
    )
    assert uneven & np.any(model.dq & pixel.JUMP_DET)
    pipeline.log.info(
        "Status of the step:             linearity     "
        + str(model.meta.cal_step.assign_wcs)
    )
    pipeline.log.info(
        "DMS90 MSG: Testing completion of linearity correction in Level 2 spectroscopic"
        " output......." + passfail(model.meta.cal_step.linearity == "COMPLETE")
    )
    assert model.meta.cal_step.linearity == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             ramp_fit      "
        + str(model.meta.cal_step.ramp_fit)
    )
    pipeline.log.info(
        "DMS90 MSG: Testing completion of ramp fitting in Level 2 spectroscopic"
        " output......." + passfail(model.meta.cal_step.ramp_fit == "COMPLETE")
    )
    assert model.meta.cal_step.ramp_fit == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             saturation    "
        + str(model.meta.cal_step.saturation)
    )
    pipeline.log.info(
        "DMS90 MSG: Testing completion of saturation detection in Level 2 spectroscopic"
        " output......." + passfail(model.meta.cal_step.saturation == "COMPLETE")
    )
    assert model.meta.cal_step.saturation == "COMPLETE"

    pipeline.log.info(
        "DMS278 MSG: Testing WFI Level 2 Data Generation - Spectroscopy in Level 2 spectroscopic"
        " output......."
        + passfail(model.meta.cal_step.dq_init == "COMPLETE")
        + passfail(model.meta.cal_step.saturation == "COMPLETE")
        + passfail(model.meta.cal_step.refpix == "COMPLETE")
        + passfail(model.meta.cal_step.linearity == "COMPLETE")
        + passfail(model.meta.cal_step.jump == "COMPLETE")
        + passfail(model.meta.cal_step.ramp_fit == "COMPLETE")
        + passfail(model.meta.cal_step.assign_wcs == "COMPLETE")
        + passfail(model.meta.cal_step.flat_field == "SKIPPED")
        + passfail(model.meta.cal_step.photom == "SKIPPED")
        + passfail(model.meta.cal_step.source_detection == "SKIPPED")
        + passfail(model.meta.cal_step.tweakreg == "SKIPPED")
    )
    assert model.meta.cal_step.dq_init == "COMPLETE"
    assert model.meta.cal_step.saturation == "COMPLETE"
    assert model.meta.cal_step.refpix == "COMPLETE"
    assert model.meta.cal_step.linearity == "COMPLETE"
    assert model.meta.cal_step.jump == "COMPLETE"
    assert model.meta.cal_step.ramp_fit == "COMPLETE"
    assert model.meta.cal_step.assign_wcs == "COMPLETE"
    assert model.meta.cal_step.flat_field == "SKIPPED"
    assert model.meta.cal_step.photom == "SKIPPED"
    assert model.meta.cal_step.source_detection == "SKIPPED"
    assert model.meta.cal_step.tweakreg == "SKIPPED"

    # DMS91 data quality tests
    pipeline.log.info(
        "DMS91 MSG: Testing existence of data quality array (dq) in Level 2"
        " spectroscopic output......." + passfail("dq" in model.keys())
    )
    assert "dq" in model.keys()
    pipeline.log.info(
        "DMS91 MSG: Testing existence of general error array (err) in Level 2"
        " spectroscopic output......." + passfail("err" in model.keys())
    )
    assert "err" in model.keys()
    pipeline.log.info(
        "DMS91 MSG: Testing existence of Poisson noise variance array (var_poisson) in"
        " Level 2 spectroscopic output.." + passfail("var_poisson" in model.keys())
    )
    assert "var_poisson" in model.keys()
    pipeline.log.info(
        "DMS91 MSG: Testing existence of read noise variance array (var_rnoise) in"
        " Level 2 spectroscopic output..." + passfail("var_rnoise" in model.keys())
    )
    assert "var_rnoise" in model.keys()

    # DMS88 total exposure time test
    pipeline.log.info(
        "DMS88 MSG: Testing existence of total exposure time (exposure_time) in Level 2"
        " spectroscopic output." + passfail("exposure_time" in model.meta.exposure)
    )
    assert "exposure_time" in model.meta.exposure

    # DMS93 WCS tests
    pipeline.log.info(
        "DMS93 MSG: Testing that the wcs boundingbox was generated......."
        + passfail(len(model.meta.wcs.bounding_box) == 2)
    )
    assert len(model.meta.wcs.bounding_box) == 2

    # Save original wcs information
    orig_wcs = copy.deepcopy(model.meta.wcs)
    del model.meta["wcs"]

    # Create new pointing for the model
    # RA & Dec are each shifted + 10 degrees, unless they are near
    # the upper limit, in which case they are shifted -10 degrees.
    delta = [10.0, 10.0]
    if model.meta.wcsinfo.ra_ref >= 350.0:
        delta[0] *= -1.0

    if model.meta.wcsinfo.dec_ref >= 80.0:
        delta[1] *= -1.0

    model.meta.wcsinfo.ra_ref += delta[0]
    model.meta.wcsinfo.dec_ref += delta[1]

    # Create new wcs object for the new pointing
    model = AssignWcsStep.call(model)

    rtdata.output = output.rsplit(".", 1)[0] + "_repoint.asdf"
    model.to_asdf(rtdata.output)

    # Test that repointed file matches truth
    rtdata.get_truth("truth/WFI/grism/" + output.rsplit(".", 1)[0] + "_repoint.asdf")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()

    pipeline.log.info(
        "DMS93 MSG: Testing that the different pointings create differing wcs......."
        + passfail(
            ((np.abs(orig_wcs(2048, 2048)[0] - model.meta.wcs(2048, 2048)[0])) - 10.0)
            < 1.0
        )
    )
    assert_allclose(
        [orig_wcs(2048, 2048)[0] + delta[0], orig_wcs(2048, 2048)[1] + delta[1]],
        model.meta.wcs(2048, 2048),
        atol=1.0,
    )


@pytest.mark.bigdata
def test_processing_pipeline_all_saturated(rtdata, ignore_asdf_paths):
    """Tests for fully saturated data skipping steps in the pipeline"""
    input_data = "r0000101001001001001_01101_0001_WFI01_ALL_SATURATED_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    # Test Pipeline
    output = "r0000101001001001001_01101_0001_WFI01_ALL_SATURATED_cal.asdf"
    rtdata.output = output
    args = [
        "--disable-crds-steppars",
        "roman_elp",
        rtdata.input,
    ]
    ExposurePipeline.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")

    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()

    # Ensure step completion is as expected
    model = rdm.open(rtdata.output)

    assert model.meta.cal_step.dq_init == "COMPLETE"
    assert model.meta.cal_step.saturation == "COMPLETE"
    assert model.meta.cal_step.linearity == "SKIPPED"
    assert model.meta.cal_step.dark == "SKIPPED"
    assert model.meta.cal_step.jump == "SKIPPED"
    assert model.meta.cal_step.ramp_fit == "SKIPPED"
    assert model.meta.cal_step.assign_wcs == "SKIPPED"
    assert model.meta.cal_step.flat_field == "SKIPPED"
    assert model.meta.cal_step.photom == "SKIPPED"



#
# Roman documentation build configuration file
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime
import importlib
import os
import sys
from distutils.version import LooseVersion
from pathlib import Path

import sphinx
import stsci_rtd_theme

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

from sphinx.ext.autodoc import AttributeDocumenter

from romancal.stpipe import RomanStep


class StepSpecDocumenter(AttributeDocumenter):
    def should_suppress_value_header(self):
        if self.name == "spec" and issubclass(self.parent, RomanStep):
            # if this attribute is named "spec" and belongs to a "Step"
            # don't show the value, it will be formatted in add_context below
            return True
        return super().should_suppress_value_header()

    def add_content(self, more_content):
        super().add_content(more_content)
        if self.name != "spec" or not issubclass(self.parent, RomanStep):
            return
        if not self.object.strip():
            return

        # format the long "Step.spec" string to improve readability
        source_name = self.get_sourcename()
        self.add_line("::", source_name, 0)
        self.add_line("  ", source_name, 1)
        txt = "\n".join(l.strip() for l in self.object.strip().splitlines())
        self.add_line(f"  {txt}", source_name, 2)


def setup(app):
    try:
        app.add_css_file("stsci.css")
    except AttributeError:
        app.add_stylesheet("stsci.css")
    # add a custom AttributeDocumenter subclass to handle Step.spec formatting
    app.add_autodocumenter(StepSpecDocumenter, True)


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath("../"))
sys.path.insert(0, os.path.abspath("romancal/"))
sys.path.insert(0, os.path.abspath("exts/"))

# -- General configuration ------------------------------------------------
with open(Path(__file__).parent.parent / "pyproject.toml", "rb") as configuration_file:
    conf = tomllib.load(configuration_file)
setup_cfg = conf["project"]

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.3'

on_rtd = os.environ.get("READTHEDOCS", None) == "True"


def check_sphinx_version(expected_version):
    sphinx_version = LooseVersion(sphinx.__version__)
    expected_version = LooseVersion(expected_version)
    if sphinx_version < expected_version:
        raise RuntimeError(
            f"At least Sphinx version {expected_version} is required to build this"
            f" documentation.  Found {sphinx_version}."
        )


# Configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/devdocs", None),
    "scipy": ("http://scipy.github.io/devdocs", None),
    "matplotlib": ("http://matplotlib.org/", None),
    "gwcs": ("https://gwcs.readthedocs.io/en/latest/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "photutils": ("https://photutils.readthedocs.io/en/stable/", None),
    "webbpsf": ("https://webbpsf.readthedocs.io/en/latest/", None),
    "roman_datamodels": ("https://roman-datamodels.readthedocs.io/en/latest/", None),
    "rad": ("https://rad.readthedocs.io/en/latest/", None),
}

intersphinx_disabled_reftypes = ["*"]

if sys.version_info[0] == 2:
    intersphinx_mapping["python"] = ("http://docs.python.org/2/", None)
    intersphinx_mapping["pythonloc"] = (
        "http://docs.python.org/",
        os.path.abspath(
            os.path.join(os.path.dirname(__file__), "local/python2_local_links.inv")
        ),
    )

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "numfig",
    "pytest_doctestplus.sphinx.doctestplus",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.automodsumm",
    "sphinx_automodapi.autodoc_enhancements",
    "sphinx_automodapi.smart_resolver",
    "sphinxcontrib.jquery",
]

if on_rtd:
    extensions.append("sphinx.ext.mathjax")

elif LooseVersion(sphinx.__version__) < LooseVersion("1.4"):
    extensions.append("sphinx.ext.pngmath")
else:
    extensions.append("sphinx.ext.imgmath")

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# A list of warning types to suppress arbitrary warning messages. We mean to
# override directives in astropy_helpers.sphinx.ext.autodoc_enhancements,
# thus need to ignore those warning. This can be removed once the patch gets
# released in upstream Sphinx (https://github.com/sphinx-doc/sphinx/pull/1843).
# Suppress the warnings requires Sphinx v1.4.2

suppress_warnings = [
    "app.add_directive",
]

# General information about the project
project = setup_cfg["name"]
author = f'{setup_cfg["authors"][0]["name"]} <{setup_cfg["authors"][0]["email"]}>'
copyright = f"{datetime.datetime.now().year}, {author}"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
package = importlib.import_module(setup_cfg["name"])
try:
    version = package.__version__.split("-", 1)[0]
    # The full version, including alpha/beta/rc tags.
    release = package.__version__
except AttributeError:
    version = "dev"
    release = "dev"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog = """.. _romancal: high-level_API.html"""

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = "obj"

# Don't show summaries of the members in each class along with the
# class' docstring
numpydoc_show_class_members = False

autosummary_generate = True

automodapi_toctreedirnm = "api"

# Class documentation should contain *both* the class docstring and
# the __init__ docstring
autoclass_content = "both"

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    "-Nfontsize=10",
    "-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Efontsize=10",
    "-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Gfontsize=10",
    "-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
]

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "default"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# Mapping for links to the ASDF Standard in ASDF schema documentation
asdf_schema_reference_mappings = [
    (
        "tag:stsci.edu:asdf",
        "http://asdf-standard.readthedocs.io/en/latest/generated/stsci.edu/asdf/",
    ),
]

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"
html_logo = "_static/roman_logo_white_w100px.png"
html_favicon = "_static/favicon.ico"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {"collapse_navigation": True, "display_version": True}
#        "nosidebar": "false",
#        "sidebarbgcolor": "#4db8ff",
#        "sidebartextcolor": "black",
#        "sidebarlinkcolor": "black",
#        "headbgcolor": "white",
#        }

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [stsci_rtd_theme.get_html_theme_path()]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%b %d, %Y"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {"**": ["globaltoc.html", "relations.html", "searchbox.html"]}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "romandoc"

# -- Options for LaTeX output ---------------------------------------------

# latex_elements = {
#    # The paper size ('letterpaper' or 'a4paper').
#    'papersize': 'letterpaper',
#    # The font size ('10pt', '11pt' or '12pt').
#    'pointsize': '14pt',
#    # Additional stuff for the LaTeX preamble.
#    'preamble': r'''\usepackage{enumitem} \setlistdepth{99}'''
# }

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
# latex_documents = [
#  ('index', 'roman.tex', u'Roman Pipeline Documentation',
#   u'roman', 'manual'),
# ]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = '_static/Romanlogocrop.png'

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = 'True'

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", "romancal", "Roman Pipeline Documentation", ["romancal"], 1)]

# If true, show URL addresses after external links.
man_show_urls = True

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "romancal",
        "Roman Pipeline Documentation",
        "romancal",
        "romancal",
        "Roman Pipeline Documentation",
        "Miscellaneous",
    ),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
texinfo_show_urls = "inline"

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False


# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = "Roman"
epub_author = "STSCI"
epub_publisher = "STSCI"
epub_copyright = "2017, AURA"

# The basename for the epub file. It defaults to the project name.
# epub_basename = u'Roman'

# The HTML theme for the epub output. Since the default themes are not
# optimized for small screen space, using the same theme for HTML and
# epub output is usually not wise. This defaults to 'epub', a theme designed
# to save visual space.
epub_theme = "epub"

# The language of the text. It defaults to the language option
# or en if the language is not set.
# epub_language = ''

# The scheme of the identifier. Typical schemes are ISBN or URL.
# epub_scheme = ''

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
# epub_identifier = ''

# A unique identification for the text.
# epub_uid = ''

# A tuple containing the cover image and cover page html template filenames.
# epub_cover = ()

# A sequence of (type, uri, title) tuples for the guide element of content.opf.
# epub_guide = ()

# HTML files that should be inserted before the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_pre_files = []

# HTML files that should be inserted after the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_post_files = []

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

# The depth of the table of contents in toc.ncx.
# epub_tocdepth = 3

# Allow duplicate toc entries.
# epub_tocdup = True

# Choose between 'default' and 'includehidden'.
# epub_tocscope = 'default'

# Fix unsupported image types using the PIL.
# epub_fix_images = False

# Scale large images.
# epub_max_image_width = 0

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# epub_show_urls = 'inline'

# If false, no index is generated.
# epub_use_index = True



"""Create an association from a list"""

import argparse
import sys

from .lib.rules_elpp_base import DMS_ELPP_Base
from .registry import AssociationRegistry

__all__ = ["asn_from_list"]


def asn_from_list(items, rule=DMS_ELPP_Base, **kwargs):
    """Create an association from a list

    Parameters
    ----------
    items: [object [, ...]]
        List of items to add.

    rule: `Association` rule
        The association rule to use.

    kwargs: dict
        Other named parameters required or pertinent to adding
        the items to the association.

    Returns
    -------
    association: `Association`-based instance
        The association with the items added.

    Notes
    -----
    This is a lower-level tool for artificially creating
    an association. As such, the association created may not be valid.
    It is presume the user knows what they are doing.
    """
    asn = rule()
    asn._add_items(items, **kwargs)
    return asn


class Main:
    """Command-line interface for list_to_asn

    Parameters
    ----------
    args: [str, ...], or None
        The command line arguments. Can be one of
            - `None`: `sys.argv` is then used.
            - `[str, ...]`: A list of strings which create the command line
              with the similar structure as `sys.argv`
    """

    def __init__(self, args=None):
        if args is None:
            args = sys.argv[1:]
        if isinstance(args, str):
            args = args.split(" ")

        parser = argparse.ArgumentParser(
            description="Create an association from a list of files",
            usage="asn_from_list -o mosaic_asn.json\n--product-name my_mosaic *.fits",
        )

        parser.add_argument(
            "-o",
            "--output-file",
            type=str,
            required=True,
            help="File to write association to",
        )

        parser.add_argument(
            "-f",
            "--format",
            type=str,
            default="json",
            help='Format of the association files. Default: "%(default)s"',
        )

        parser.add_argument(
            "--product-name",
            type=str,
            help="The product name when creating a Level 3 association",
        )

        parser.add_argument(
            "-r",
            "--rule",
            type=str,
            default="DMS_ELPP_Base",
            help=(
                'The rule to base the association structure on. Default: "%(default)s"'
            ),
        )
        parser.add_argument(
            "--ruledefs",
            action="append",
            help=(
                "Association rules definition file(s) If not specified, the default"
                " rules will be searched."
            ),
        )
        parser.add_argument(
            "-i",
            "--id",
            type=str,
            default="o999",
            help='The association candidate id to use. Default: "%(default)s"',
            dest="acid",
        )

        parser.add_argument(
            "filelist",
            type=str,
            nargs="+",
            help="File list to include in the association",
        )

        parsed = parser.parse_args(args=args)
        print("Parsed args:", parsed)

        # Get the rule
        rule = AssociationRegistry(parsed.ruledefs, include_bases=True)[parsed.rule]

        with open(parsed.output_file, "w") as outfile:
            asn = asn_from_list(
                parsed.filelist,
                rule=rule,
                product_name=parsed.product_name,
                acid=parsed.acid,
            )
            _, serialized = asn.dump(format=parsed.format)
            outfile.write(serialized)



"""
This module the parts of the provided reference code that this step is based on,
which are relevant to the reference pixel correction. The code is mildly modified
so that it can be split into smaller more testable pieces to aid in verification
of the current implementation.

This module and the tests based off this module are not intended to be permanent.
Instead, this module is intended to demonstrate the hand-off between the provided
reference code and the current implementation. Once the current implementation is
accepted, this module will be removed and the tests will be updated to use regression
data generated by the current implementation.

This way there is a clear unquestionable demonstration of the hand-off between the
provided reference code and this implementation contained within the git history.
of the project itself.
"""

import threading

import numpy as np
import scipy.fft as spfft

NUM_OUTPUT_CHANS = 33
NUM_COLS_PER_OUTPUT_CHAN = 128
NUM_COLS = NUM_OUTPUT_CHANS * NUM_COLS_PER_OUTPUT_CHAN
END_OF_ROW_PIXEL_PAD = 12
NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD = NUM_COLS_PER_OUTPUT_CHAN + END_OF_ROW_PIXEL_PAD
ALL_CHAN_RANGE = range(NUM_OUTPUT_CHANS)
REFERENCE_CHAN = 32
REFPIX_NORM = NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD / 4.0
CHAN_RANGE_WITHOUT_REFERENCES = range(NUM_OUTPUT_CHANS - 1)


def remove_linear_trends(data_FramesRowsCols, subtractOffsetOnly):
    """
    Remove linear trends (gains/slope, biases/offsets) per pixel
    :param data_FramesRowsCols: image data [numFrames][numRows][numCols].
        The data is modified IN PLACE
    :param subtractOffsetOnly: if True, only the liner model offset is removed.
        If False, the offset and slope are removed
    :return per-pixel image array of modeled m (slope) and b (offset)
    """

    numFrames = len(data_FramesRowsCols)
    numRows = len(data_FramesRowsCols[0])
    numCols = len(data_FramesRowsCols[0][0])

    # for mat multiply
    data_FramesRowsCols = data_FramesRowsCols.reshape((numFrames, numCols * numRows))

    frameRamp = np.arange(numFrames, dtype=data_FramesRowsCols.dtype)

    # subtract mean(frameRampMinusMean) to minimize correlation of slope and offset
    frameRampMinusMean = frameRamp - np.mean(frameRamp)

    sxy = np.matmul(
        frameRampMinusMean, data_FramesRowsCols, dtype=data_FramesRowsCols.dtype
    )
    sx = np.sum(frameRampMinusMean)
    sxx = np.sum(frameRampMinusMean**2)
    sy = np.sum(data_FramesRowsCols, axis=0)

    m = (numFrames * sxy - sx * sy) / (numFrames * sxx - sx**2)
    b = (sy * sxx - sxy * sx) / (numFrames * sxx - sx**2)

    for frame in range(numFrames):
        if not subtractOffsetOnly:
            data_FramesRowsCols[frame, :] -= frameRampMinusMean[frame] * m + b
        else:
            data_FramesRowsCols[frame, :] -= b  # subtract offset

    data_FramesRowsCols = data_FramesRowsCols.reshape(numFrames, numRows, numCols)
    return m.reshape(numRows, numCols), b.reshape(numRows, numCols)


def aligned_channels(data):
    numFrames = data.shape[0]
    numRows = data.shape[1]

    data_chansFramesRowsPhyscols = np.transpose(
        data.reshape((numFrames, numRows, NUM_OUTPUT_CHANS, NUM_COLS_PER_OUTPUT_CHAN)),
        (2, 0, 1, 3),
    )

    for chanNum in range(1, NUM_OUTPUT_CHANS, 2):
        data_chansFramesRowsPhyscols[chanNum, :, :, :] = data_chansFramesRowsPhyscols[
            chanNum, :, :, ::-1
        ]

    return np.pad(
        data_chansFramesRowsPhyscols,
        ((0, 0), (0, 0), (0, 0), (0, END_OF_ROW_PIXEL_PAD)),
    )


def remove_linear_trends_per_frame(
    data_chansFramesRowsChancols, subtractOffsetOnly, multiThread=True
):
    """
    Entry point for Fitting and removal of slopes per frame to remove issues
        at frame boundaries.
    :param data_chansFramesRowsChancols:
        [numOutChannels(33)][numFrames][numRows][numColsWithPadding].
        The data is modified IN PLACE
    :param subtractOffsetOnly: if True, only the liner model offset is removed.
        If False, the offset and slope are removed
    """

    numFrames = len(data_chansFramesRowsChancols[0])
    numRows = len(data_chansFramesRowsChancols[0][0])
    timeDomainRange = np.arange(
        NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD * numRows,
        dtype=data_chansFramesRowsChancols.dtype,
    ).reshape(numRows, NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD)
    timeDomainRange = timeDomainRange - timeDomainRange.mean()

    exec_channel_func_threads(
        ALL_CHAN_RANGE,
        remove_linear_trends_per_frame_chan_func,
        (numFrames, timeDomainRange, data_chansFramesRowsChancols, subtractOffsetOnly),
        multiThread=multiThread,
    )


def remove_linear_trends_per_frame_chan_func(
    chanNumber,
    numFrames,
    timeDomainRange,
    data_chansFramesRowsChancols,
    subtractOffsetOnly,
):
    """
    Per-channel function for fitting and removal of slopes per frame to remove
        issues at frame boundaries - single channel only
    :param chanNumber:
    :param numFrames:
    :param timeDomainRange:
    :param data_chansFramesRowsChancols:
        [numOutChannels(33)][numFrames][numRows][numColsWithPadding].
        The data is modified IN PLACE
    :param subtractOffsetOnly: if True, only the liner model offset is removed.
        If False, the offset and slope are removed
    """

    numRows = len(data_chansFramesRowsChancols[0][0])
    REFERENCE_ROWS = list(range(0, 4)) + list(range(numRows - 4, numRows))

    ab_3 = np.zeros(
        (NUM_OUTPUT_CHANS, numFrames, 2), dtype=np.float64
    )  # for top+bottom ref pixel rows
    tdRef = timeDomainRange[REFERENCE_ROWS, :]

    for frame in range(numFrames):
        xVals = tdRef[
            (data_chansFramesRowsChancols[chanNumber, frame, REFERENCE_ROWS, :] != 0)
        ]  # where data[chanNumber,frame,row4plus4,:] != 0)
        data2 = data_chansFramesRowsChancols[chanNumber, frame, REFERENCE_ROWS, :]
        yVals = data2[(data2 != 0)]
        if xVals.size < 1 or yVals.size < 1:
            continue

        ab_3[chanNumber, frame, :] = np.polyfit(xVals, yVals, 1)

        dataNotZero = data_chansFramesRowsChancols[chanNumber, frame, :, :] != 0

        if not subtractOffsetOnly:
            data2 = (
                timeDomainRange * ab_3[chanNumber, frame, 0]
                + ab_3[chanNumber, frame, 1]
            )
        else:
            data2 = ab_3[chanNumber, frame, 1]

        data_chansFramesRowsChancols[chanNumber, frame, :, :] -= data2 * dataNotZero


def exec_channel_func_threads(chanIndexRange, targetFunc, funcArgs, multiThread=False):
    """
    Execute a function over a range of channels.  If multiThread is True, all executions
    will be executed on individual threads.

    :param chanIndexRange: E.g., 'range(NUM_OUTPUT_CHANS')
    :param targetFunc: function to call, first arg must be channel number
    :param funcArgs: function argument tuple NOT INCLUDING channel number e.g.
        (sig_data, goodPixelsOneIsMask_RowCol)
    :param multiThread: if True allocate a separate thread for each channel;
        otherwise the channels are processed sequentially in a single thread

    """

    if multiThread:
        threadList = []
        for c in chanIndexRange:
            funcArgsWithChan = (c,) + funcArgs
            threadList.append(
                threading.Thread(target=targetFunc, args=funcArgsWithChan)
            )

        for t in threadList:
            t.start()

        for t in threadList:
            t.join()
    else:
        for c in chanIndexRange:
            funcArgsWithChan = (c,) + funcArgs
            targetFunc(*funcArgsWithChan)


def getTrigInterpolationFunction(data):
    return np.sin(np.arange(NUM_OUTPUT_CHANS, dtype=data.dtype) / 32 * np.pi)[1:]


def interp_zeros_channel_fun(
    chanNumber,
    interpFunc,
    data_ChansFramesRowsCols,
    numFrames,
    numRows,
    numColsPerOutputChanWithPad,
):
    """
    Convolve interpFunc across pixels with values of 0 for a single channel.
    We are not using the outlier mask and thus possibly interpolating pixels
    with original values of 0. This is to ensure we don't have discontinuities
    (around 0) when doing the FFT interpolation, which could damage the performance.

    :param chanNumber:
    :param interpFunc:
    :param data_ChansFramesRowsCols:  Data to interpolate, which is done IN PLACE
    :param numFrames:
    :param numRows:
    :param numColsPerOutputChanWithPad:
    """
    for frame in range(numFrames):
        dat = np.reshape(
            data_ChansFramesRowsCols[chanNumber, frame, :, :],
            numColsPerOutputChanWithPad * numRows,
        )

        # Apply to areas with 0 value.
        applyMask = np.where(dat != 0, 1, 0)

        dataConvolve = np.convolve(
            dat, interpFunc, mode="same"
        )  # 'same' -> /edge_wrap in IDL
        applyMaskConvolve = np.convolve(applyMask, interpFunc, mode="same")

        sfm3 = dataConvolve / applyMaskConvolve
        dat = np.reshape(sfm3, (numRows, numColsPerOutputChanWithPad))
        mask2D = np.reshape(applyMask, (numRows, numColsPerOutputChanWithPad))
        data_ChansFramesRowsCols[chanNumber, frame, :, :] += dat * (1 - mask2D)
        # Repair any NaNs
        np.nan_to_num(data_ChansFramesRowsCols[chanNumber, :, :, :], copy=False)


def cos_interp_reference(dataUniformTime, numFrames):
    numRows = dataUniformTime.shape[2]
    interp_zeros_channel_fun(
        REFERENCE_CHAN,
        getTrigInterpolationFunction(dataUniformTime),
        dataUniformTime,
        numFrames,
        numRows,
        NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD,
    )


def getFFTApodizeFunction(numRows):
    apo_len = numRows * NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD
    apo2 = np.abs(np.fft.rfftfreq(apo_len, 1 / apo_len))
    apodize_func = np.cos(2 * np.pi * apo2 / apo_len) / 2.0 + 0.5
    return apodize_func


def fft_interp(
    chanData_FramesFlat,
    readOnlyPixelsAreOneMask_Flattenedimage,
    apodizeFilter,
    numIterations,
):
    """
    Perform FFT interpolation on pixels not marked as read-only
    :param chanData_FramesFlat: Multi-frame, single channel data with each
        frame flattened to 1D.  WILL BE UPDATED IN PLACE
    :param readOnlyPixelsAreOneMask_Flattenedimage:  Flattened image size mask
        indicating (good) pixels that should be used in the interpolation
        but not modified
    :param numFrames: number of frames in chanData_FramesFlat
    :param apodizeFilter:
    :param numIterations: number of times to iterate the interpolation
    """

    numFrames = len(chanData_FramesFlat)
    readOnlyPixelsIndices_Flat = np.where(readOnlyPixelsAreOneMask_Flattenedimage)[0]

    for frame in range(numFrames):
        chanFrameData_Flat = chanData_FramesFlat[frame]
        readOnlyPixelsValues_Flat = chanFrameData_Flat[readOnlyPixelsIndices_Flat]

        for _ in range(numIterations):
            fftResult = (
                apodizeFilter
                * spfft.rfft(chanFrameData_Flat, workers=1)
                / chanFrameData_Flat.size
            )
            # astype is added so that dtype is preserved through the computation
            # this reduces the number of copies made, and does not affect the result
            # meaningfully
            chanFrameData_Flat = spfft.irfft(
                fftResult * chanFrameData_Flat.size,
                workers=1,
                # )
            ).astype(chanFrameData_Flat.dtype)
            # Return read only pixels
            chanFrameData_Flat[readOnlyPixelsIndices_Flat] = readOnlyPixelsValues_Flat

        chanData_FramesFlat[frame, :] = chanFrameData_Flat


def fft_interp_amp33(dataUniformTime, numFrames):
    numRows = dataUniformTime.shape[2]

    refOutputOnesMask = np.zeros(
        (numRows, NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD), dtype=bool
    )
    refOutputOnesMask[:, :NUM_COLS_PER_OUTPUT_CHAN] = 1

    dataReferenceChan_FramesFlat = np.reshape(
        dataUniformTime[REFERENCE_CHAN, :, :, :],
        (numFrames, (numRows * NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD)),
    ).astype(dtype=dataUniformTime.dtype)
    fft_interp(
        dataReferenceChan_FramesFlat,
        refOutputOnesMask.flatten(),
        getFFTApodizeFunction(numRows),
        3,
    )

    return dataReferenceChan_FramesFlat.reshape(
        numFrames, numRows, NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD
    )


def left(dataUniformTime):
    l = np.copy(dataUniformTime[0, :, :, :])
    l[:, :, 4:] = 0
    return l.copy()


def left_fft(numFrames, dataUniformTime):
    numRows = dataUniformTime.shape[2]
    l = left(dataUniformTime)
    l = l.reshape((numFrames, NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD * numRows))
    return REFPIX_NORM * spfft.rfft(l / l[0].size).copy()


def right(dataUniformTime):
    r = np.copy(dataUniformTime[31, :, :, :])
    r[:, :, 4:] = 0
    return r.copy()


def right_fft(numFrames, dataUniformTime):
    numRows = dataUniformTime.shape[2]
    r = right(dataUniformTime)
    r = r.reshape((numFrames, NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD * numRows))
    return REFPIX_NORM * spfft.rfft(r / r[0].size).copy()


def amp33_fft(numFrames, dataUniformTime):
    numRows = dataUniformTime.shape[2]
    a = np.copy(dataUniformTime[REFERENCE_CHAN, :, :, :])
    a = a.reshape((numFrames, NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD * numRows))
    return spfft.rfft(a / a[0].size).copy()


def forward_fft(numFrames, dataUniformTime):
    return (
        left_fft(numFrames, dataUniformTime),
        right_fft(numFrames, dataUniformTime),
        amp33_fft(numFrames, dataUniformTime),
    )


def _corr_chan_func(
    chanNum: int,
    alpha: np.ndarray,
    gamma: np.ndarray,
    zeta: np.ndarray,
    numFrames: int,
    l: np.ndarray,
    r: np.ndarray,
    a: np.ndarray,
    corrNoise: np.ndarray,
):
    """
    Estimate the correlated noise for the input file across all frames for
        specified channel

    :param chanNum: channel (column) number
    :param alpha: weights for reference output
    :param gamma: weights for left reference columns
    :param zeta: weights for right reference columns
    :param numFrames: number of frames in data
    :param l: input file left reference column data
    :param r: input file right reference column data
    :param a: input file reference channel data
    :param corrNoise: CHANGED IN PLACE correlated noise solution for this channel
    """
    lCopy = np.copy(l).astype(gamma.dtype)
    gammaChan = gamma[chanNum, :]
    for frameNum in range(numFrames):
        lCopy[frameNum, :] *= gammaChan

    zetaChan = zeta[chanNum, :]
    for frameNum in range(numFrames):
        lCopy[frameNum, :] += r[frameNum, :] * zetaChan

    alphaChan = alpha[chanNum, :]
    for frameNum in range(numFrames):
        lCopy[frameNum, :] += a[frameNum, :] * alphaChan

    # When dealing with real-only weight streams the
    # reverse FFT transform results in half the power so
    # we need to add a factor of 2 to the normalization
    normalizationFactor = lCopy[0].size * 2
    lCopy = spfft.irfft(lCopy * normalizationFactor)

    corrNoise[chanNum, :, :] = lCopy[:]


def compute_correction(numRows, alpha, gamma, zeta, numFrames, l, r, a):
    correlatedNoise = np.zeros(
        (NUM_OUTPUT_CHANS, numFrames, numRows * NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD),
    )
    exec_channel_func_threads(
        CHAN_RANGE_WITHOUT_REFERENCES,
        _corr_chan_func,
        (alpha, gamma, zeta, numFrames, l, r, a, correlatedNoise),
        multiThread=False,
    )

    return correlatedNoise


def correction(numFrames, dataUniformTime, alpha, gamma, zeta):
    numRows = dataUniformTime.shape[2]
    l, r, a = forward_fft(numFrames, dataUniformTime)

    correlatedNoise = compute_correction(
        numRows, alpha, gamma, zeta, numFrames, l, r, a
    )

    # reshape and just take real part
    correlatedNoise = correlatedNoise.real.reshape(
        (NUM_OUTPUT_CHANS, numFrames, numRows, NUM_COLS_PER_OUTPUT_CHAN_WITH_PAD)
    )

    return correlatedNoise


def apply_correction_to_channels(numFrames, dataUniformTime, alpha, gamma, zeta):
    correlatedNoise = correction(numFrames, dataUniformTime, alpha, gamma, zeta)

    correlatedNoise = correlatedNoise[:, :, :, 0:NUM_COLS_PER_OUTPUT_CHAN].astype(
        dataUniformTime.dtype
    )

    data0 = dataUniformTime[:, :, :, :NUM_COLS_PER_OUTPUT_CHAN]
    data0 -= correlatedNoise

    return data0


def apply_correction_to_data(numFrames, dataUniformTime, alpha, gamma, zeta):
    numRows = dataUniformTime.shape[2]
    data0 = apply_correction_to_channels(numFrames, dataUniformTime, alpha, gamma, zeta)

    for i in range(1, NUM_OUTPUT_CHANS - 1, 2):
        data0[i, :, :, :] = data0[i, :, :, ::-1]

    data0 = np.transpose(data0, (1, 2, 0, 3))
    data0 = np.reshape(
        data0[:, :, CHAN_RANGE_WITHOUT_REFERENCES, :],
        (numFrames, numRows, NUM_COLS - NUM_COLS_PER_OUTPUT_CHAN),
    )

    return data0


def restore_offsets(numFrames, data0, b):
    for frame in range(numFrames):
        data0[frame, :, :] += b[:, :4096]


def apply_correction(data, alpha, gamma, zeta):
    # Remove offset
    _, b = remove_linear_trends(data, True)

    # Align channels
    dataUniformTime = aligned_channels(data)

    # Remove boundary trends
    remove_linear_trends_per_frame(dataUniformTime, False, True)

    # Perform cosine weighted interpolation
    cos_interp_reference(dataUniformTime, dataUniformTime.shape[1])

    # Perform fft interpolation
    dataUniformTime[REFERENCE_CHAN, :, :, :] = fft_interp_amp33(
        dataUniformTime, dataUniformTime.shape[1]
    )

    # Perform the correction
    data0 = apply_correction_to_data(
        dataUniformTime.shape[1], dataUniformTime, alpha, gamma, zeta
    )

    # Restore offsets
    restore_offsets(data0.shape[0], data0, b)

    return data0



import os

import requests
from astropy import table
from astropy import units as u
from astropy.coordinates import SkyCoord

from ..assign_wcs import utils as wcsutil
from ..resample import resample_utils

ASTROMETRIC_CAT_ENVVAR = "ASTROMETRIC_CATALOG_URL"
DEF_CAT_URL = "http://gsss.stsci.edu/webservices"

# VO request timeout (in seconds)
TIMEOUT = 30.0

if ASTROMETRIC_CAT_ENVVAR in os.environ:
    SERVICELOCATION = os.environ[ASTROMETRIC_CAT_ENVVAR]
else:
    SERVICELOCATION = DEF_CAT_URL

"""

Primary function for creating an astrometric reference catalog.

"""


def create_astrometric_catalog(
    input_models,
    catalog="GAIADR3",
    output="ref_cat.ecsv",
    gaia_only=False,
    table_format="ascii.ecsv",
    existing_wcs=None,
    num_sources=None,
    epoch=None,
):
    """Create an astrometric catalog that covers the inputs' field-of-view.

    Parameters
    ----------
    input_models : str, list
        Filenames of images to be aligned to astrometric catalog

    catalog : str, optional
        Name of catalog to extract astrometric positions for sources in the
        input images' field-of-view. Default: GAIADR3. Options available are
        documented on the catalog web page.

    output : str, optional
        Filename to give to the astrometric catalog read in from the master
        catalog web service.  If None, no file will be written out.

    gaia_only : bool, optional
        Specify whether or not to only use sources from GAIA in output catalog

    table_format: str, optional
        Format to be used when writing the results to a file using the `output` option.
        A full list of the options can be found here:
        https://docs.astropy.org/en/stable/io/unified.html#built-in-readers-writers

    existing_wcs : model
        existing WCS object specified by the user as generated by
        `resample.resample_utils.make_output_wcs`

    num_sources : int
        Maximum number of brightest/faintest sources to return in catalog.
        If `num_sources` is negative, return that number of the faintest
        sources.  By default, all sources are returned.

    epoch: float or str, optional
        Reference epoch used to update the coordinates for proper motion
        (in decimal year). If `None` then the epoch is obtained from
        the metadata.

    Notes
    -----
    This function will point to astrometric catalog web service defined
    through the use of the `ASTROMETRIC_CATALOG_URL` environment variable.
    Also, the default catalog to be used is set by the `DEF_CAT` variable.

    Returns
    -------
    ref_table : `~astropy.table.Table`
        Astropy Table object of the catalog

    """

    # start by creating a composite field-of-view for all inputs
    # This default output WCS will have the same plate-scale and orientation
    # as the first member in the list.
    # Fortunately, for alignment, this doesn't matter since no resampling of
    # data will be performed.
    if existing_wcs is not None:
        outwcs = existing_wcs
    else:
        outwcs = resample_utils.make_output_wcs(input_models)
    radius, fiducial = compute_radius(outwcs)

    # perform query for this field-of-view
    epoch = (
        epoch
        if epoch is not None
        else input_models[0].meta.exposure.mid_time.decimalyear
    )
    if not isinstance(epoch, float):
        # keep only decimal point and digit characters
        epoch = float("".join(c for c in str(epoch) if c == "." or c.isdigit()))

    ref_dict = get_catalog(
        fiducial[0], fiducial[1], epoch=epoch, sr=radius, catalog=catalog
    )
    colnames = ("ra", "dec", "mag", "objID", "epoch")

    ref_table = ref_dict[colnames]

    # Add catalog name as meta data
    ref_table.meta["catalog"] = catalog
    ref_table.meta["gaia_only"] = gaia_only

    # rename coordinate columns to be consistent with tweakwcs
    ref_table.rename_column("ra", "RA")
    ref_table.rename_column("dec", "DEC")

    # Append GAIA ID as a new column to the table...
    gaia_sources = []
    for source in ref_dict:
        if "GAIAsourceID" in source:
            g = source["GAIAsourceID"]
            if gaia_only and g.strip() == "":
                continue
        else:
            g = "-1"  # indicator for no source ID extracted
        gaia_sources.append(g)

    gaia_col = table.Column(data=gaia_sources, name="GaiaID", dtype="U25")
    ref_table.add_column(gaia_col)

    # sort table by magnitude, fainter to brightest
    ref_table.sort("mag", reverse=True)

    # If specified by the use through the 'num_sources' parameter,
    # trim the returned catalog down to just the brightest 'num_sources' sources
    # Should 'num_sources' be a negative value, it will return the faintest
    # 'num_sources' sources.
    if num_sources is not None:
        indx = -1 * num_sources
        ref_table = ref_table[:indx] if num_sources < 0 else ref_table[indx:]

    # Write out table to a file, if specified
    if output is not None:
        ref_table.write(output, format=table_format, overwrite=True)

    return ref_table


"""

Utility functions for creating an astrometric reference catalog.

"""


def compute_radius(wcs):
    """Compute the radius from the center to the furthest edge of the WCS."""

    fiducial = wcsutil.compute_fiducial([wcs], wcs.bounding_box)
    img_center = SkyCoord(ra=fiducial[0] * u.degree, dec=fiducial[1] * u.degree)
    wcs_foot = wcs.footprint()
    img_corners = SkyCoord(ra=wcs_foot[:, 0] * u.degree, dec=wcs_foot[:, 1] * u.degree)
    radius = img_center.separation(img_corners).max().value

    return radius, fiducial


def get_catalog(ra, dec, epoch=2016.0, sr=0.1, catalog="GAIADR3", timeout=TIMEOUT):
    """Extract catalog from VO web service.

    Parameters
    ----------
    ra : float
        Right Ascension (RA) of center of field-of-view (in decimal degrees)

    dec : float
        Declination (Dec) of center of field-of-view (in decimal degrees)

    epoch: float, optional
        Reference epoch used to update the coordinates for proper motion
        (in decimal year). Default: 2016.0.

    sr : float, optional
        Search radius (in decimal degrees) from field-of-view center to use
        for sources from catalog.  Default: 0.1 degrees

    catalog : str, optional
        Name of catalog to query, as defined by web-service.  Default: 'GAIADR3'

    timeout : float, optional
        Set the request timeout (in seconds). Default: 30 s.

    Returns
    -------
        A Table object of returned sources with all columns as provided by catalog.

    """
    service_type = "vo/CatalogSearch.aspx"
    spec_str = "RA={}&DEC={}&EPOCH={}&SR={}&FORMAT={}&CAT={}&MINDET=5"
    headers = {"Content-Type": "text/csv"}
    fmt = "CSV"

    spec = spec_str.format(ra, dec, epoch, sr, fmt, catalog)
    service_url = f"{SERVICELOCATION}/{service_type}?{spec}"
    try:
        rawcat = requests.get(service_url, headers=headers, timeout=timeout)
    except requests.exceptions.ConnectionError:
        raise requests.exceptions.ConnectionError(
            "Could not connect to the VO API server. Try again later."
        )
    except requests.exceptions.Timeout:
        raise requests.exceptions.Timeout("The request to the VO API server timed out.")
    except requests.exceptions.RequestException:
        raise requests.exceptions.RequestException(
            "There was an unexpected error with the request."
        )
    # convert from bytes to a String
    r_contents = rawcat.content.decode()
    rstr = r_contents.split("\r\n")
    # remove initial line describing the number of sources returned
    # CRITICAL to proper interpretation of CSV data
    del rstr[0]
    if len(rstr) == 0:
        raise Exception(
            """VO catalog service returned no results.\n
            Hint: maybe reviewing the search parameters might help."""
        )

    return table.Table.read(rstr, format="csv")



#! /usr/bin/env python

import roman_datamodels as rdm

from romancal.photom import photom
from romancal.stpipe import RomanStep

__all__ = ["PhotomStep"]


class PhotomStep(RomanStep):
    """
    PhotomStep: Module for loading photometric conversion information from
        reference files and attaching to the input science data model
    """

    reference_file_types = ["photom"]

    def process(self, input):
        """Perform the photometric calibration step

        Parameters
        ----------
        input : Roman level 2 image datamodel (wfi_image-1.x.x)
            input roman datamodel

        Returns
        -------
        output_model : Roman level 2 image datamodel (wfi_image-1.x.x)
            output roman datamodel
        """

        # Open the input data model
        with rdm.open(input, lazy_load=False) as input_model:
            # Get reference file
            reffile = self.get_reference_file(input_model, "photom")

            # Open the relevant photom reference file as a datamodel
            if reffile is not None and reffile != "N/A":
                # If there is a reference file, perform photom application
                photom_model = rdm.open(reffile)
                self.log.debug(f"Using PHOTOM ref file: {reffile}")

                # Do the correction
                if input_model.meta.exposure.type == "WFI_IMAGE":
                    output_model = photom.apply_photom(input_model, photom_model)
                    output_model.meta.cal_step.photom = "COMPLETE"
                else:
                    self.log.warning("No photometric corrections for spectral data")
                    self.log.warning("Photom step will be skipped")
                    input_model.meta.cal_step.photom = "SKIPPED"
                    input_model.meta.photometry.pixelarea_arcsecsq = None
                    input_model.meta.photometry.pixelarea_steradians = None
                    input_model.meta.photometry.conversion_megajanskys = None
                    input_model.meta.photometry.conversion_microjanskys = None
                    input_model.meta.photometry.conversion_megajanskys_uncertainty = (
                        None
                    )
                    input_model.meta.photometry.conversion_microjanskys_uncertainty = (
                        None
                    )
                    output_model = input_model
                photom_model.close()

            else:
                # Skip Photom step if no photom file
                self.log.warning("No PHOTOM reference file found")
                self.log.warning("Photom step will be skipped")
                input_model.meta.cal_step.photom = "SKIPPED"
                output_model = input_model

        if self.save_results:
            try:
                self.suffix = "photom"
            except AttributeError:
                self["suffix"] = "photom"

        return output_model



"""
Roman step for sky matching.
"""

import logging
from copy import deepcopy
from itertools import chain

import numpy as np
from astropy.nddata.bitmask import bitfield_to_boolean_mask, interpret_bit_flags
from roman_datamodels import datamodels as rdd
from roman_datamodels.dqflags import pixel

from romancal.datamodels import ModelContainer
from romancal.stpipe import RomanStep

from .skyimage import SkyImage
from .skymatch import match
from .skystatistics import SkyStats

__all__ = ["SkyMatchStep"]


class SkyMatchStep(RomanStep):
    """
    SkyMatchStep: Subtraction or equalization of sky background in science images.
    """

    class_alias = "skymatch"

    spec = """
        # General sky matching parameters:
        skymethod = option('local', 'global', 'match', 'global+match', default='match') # sky computation method
        match_down = boolean(default=True) # adjust sky to lowest measured value?
        subtract = boolean(default=False) # subtract computed sky from image data?

        # Image's bounding polygon parameters:
        stepsize = integer(default=None) # Max vertex separation

        # Sky statistics parameters:
        skystat = option('median', 'midpt', 'mean', 'mode', default='mode') # sky statistics
        dqbits = string(default='~DO_NOT_USE+NON_SCIENCE') # "good" DQ bits
        lower = float(default=None) # Lower limit of "good" pixel values
        upper = float(default=None) # Upper limit of "good" pixel values
        nclip = integer(min=0, default=5) # number of sky clipping iterations
        lsigma = float(min=0.0, default=4.0) # Lower clipping limit, in sigma
        usigma = float(min=0.0, default=4.0) # Upper clipping limit, in sigma
        binwidth = float(min=0.0, default=0.1) # Bin width for 'mode' and 'midpt' `skystat`, in sigma
    """  # noqa: E501

    reference_file_types = []

    def process(self, input):
        self.log.setLevel(logging.DEBUG)
        self._is_asn = False

        img = ModelContainer(
            input, save_open=not self._is_asn, return_open=not self._is_asn
        )

        self._dqbits = interpret_bit_flags(self.dqbits, flag_name_map=pixel)

        # set sky statistics:
        self._skystat = SkyStats(
            skystat=self.skystat,
            lower=self.lower,
            upper=self.upper,
            nclip=self.nclip,
            lsig=self.lsigma,
            usig=self.usigma,
            binwidth=self.binwidth,
        )

        # group images by their "group id":
        grp_img = chain.from_iterable(img.models_grouped)

        # create a list of "Sky" Images and/or Groups:
        images = [self._imodel2skyim(g) for grp_id, g in enumerate(grp_img, start=1)]

        # match/compute sky values:
        match(
            images,
            skymethod=self.skymethod,
            match_down=self.match_down,
            subtract=self.subtract,
        )

        # set sky background value in each image's meta:
        for im in images:
            if isinstance(im, SkyImage):
                self._set_sky_background(
                    im, "COMPLETE" if im.is_sky_valid else "SKIPPED"
                )
            else:
                for gim in im:
                    self._set_sky_background(
                        gim, "COMPLETE" if gim.is_sky_valid else "SKIPPED"
                    )

        return ModelContainer([x.meta["image_model"] for x in images])

    def _imodel2skyim(self, image_model):
        input_image_model = image_model

        if "background" not in image_model.meta:
            # TODO: remove this block when ``rad`` has a background schema.
            # This is a temporary workaround to insert a 'background'
            # entry into the metadata, which we'll later put into ``rad``:
            image_model.meta["background"] = dict(
                level=None, subtracted=None, method=None
            )
        if self._is_asn:
            image_model = rdd.open(image_model)

        if self._dqbits is None:
            dqmask = np.isfinite(image_model.data).astype(dtype=np.uint8)
        else:
            dqmask = bitfield_to_boolean_mask(
                image_model.dq, self._dqbits, good_mask_value=1, dtype=np.uint8
            ) * np.isfinite(image_model.data)

        # see if 'skymatch' was previously run and raise an exception
        # if 'subtract' mode has changed compared to the previous pass:
        level = image_model.meta["background"]["level"]
        if image_model.meta["background"]["subtracted"] is None:
            if level is not None:
                if self._is_asn:
                    image_model.close()

                # report inconsistency:
                raise ValueError(
                    "Background level was set but the "
                    "'subtracted' property is undefined (None)."
                )
            # Assume level is zero:
            level = 0.0

        else:
            if image_model.meta["background"]["subtracted"] and level is None:
                # NOTE: In principle we could assume that level is 0 and
                # possibly add a log entry documenting this, however,
                # at this moment I think it is safer to quit and...
                #
                # report inconsistency:
                if self._is_asn:
                    image_model.close()

                raise ValueError(
                    "Background level was subtracted but the "
                    "'level' property is undefined (None)."
                )

            if image_model.meta["background"]["subtracted"] and self.subtract:
                # cannot run 'skymatch' step on already "skymatched" images
                # when 'subtract' spec is inconsistent with
                # meta.background.subtracted:
                if self._is_asn:
                    image_model.close()

                raise ValueError(
                    "'subtract' step's specification is "
                    "inconsistent with background info already "
                    "present in image '{:s}' meta.".format(image_model.meta.filename)
                )

        wcs = deepcopy(image_model.meta.wcs)

        sky_im = SkyImage(
            image=image_model.data,
            wcs_fwd=wcs.forward_transform,
            wcs_inv=wcs.backward_transform,
            pix_area=1.0,  # TODO: pixel area
            convf=1.0,  # TODO: conv. factor to brightness
            mask=dqmask,
            id=image_model.meta.filename,  # file name?
            skystat=self._skystat,
            stepsize=self.stepsize,
            reduce_memory_usage=self._is_asn,
            meta={"image_model": input_image_model},
        )

        if self._is_asn:
            image_model.close()

        if self.subtract:
            sky_im.sky = level

        return sky_im

    def _set_sky_background(self, sky_image, step_status):
        image = sky_image.meta["image_model"]
        sky = sky_image.sky

        if self._is_asn:
            dm = rdd.open(image)
        else:
            dm = image

        if step_status == "COMPLETE":
            # TODO: remove this block when ``rad`` has a background schema:
            # https://github.com/spacetelescope/rad/issues/247
            # This is a temporary workaround to access a 'background'
            # entry into metadata as a Python dict, which we'll later define with
            # a schema in ``rad``:
            dm.meta["background"]["method"] = str(self.skymethod)
            dm.meta["background"]["level"] = sky
            dm.meta["background"]["subtracted"] = (
                self.subtract or dm.meta["background"]["subtracted"]
            )

            if self.subtract:
                dm.data[...] = sky_image.image[...]

        dm.meta.cal_step.skymatch = step_status

        if self._is_asn:
            dm.save(image)
            dm.close()



"""
 Unit tests for the Roman source detection step code
"""

import os
import tempfile
from copy import deepcopy

import numpy as np
import pytest
from astropy import units as u
from photutils.psf import PSFPhotometry
from roman_datamodels import maker_utils as testutil
from roman_datamodels.datamodels import ImageModel

from romancal.lib.psf import create_gridded_psf_model, fit_psf_to_image_model

n_trials = 15
image_model_shape = (50, 50)
rng = np.random.default_rng(0)


def setup_inputs(
    shape=image_model_shape,
    noise=1.0,
    seed=None,
):
    """
    Return ImageModel of level 2 image.
    """
    wfi_image = testutil.mk_level2_image(shape=shape)
    wfi_image.data = u.Quantity(
        np.ones(shape, dtype=np.float32), u.DN / u.s, dtype=np.float32
    )
    wfi_image.meta.filename = "filename"
    wfi_image.meta.instrument["optical_element"] = "F087"

    # add noise to data
    if noise is not None:
        setup_rng = np.random.default_rng(seed or 19)
        wfi_image.data = u.Quantity(
            setup_rng.normal(scale=noise, size=shape),
            u.DN / u.s,
            dtype=np.float32,
        )
        wfi_image.err = noise * np.ones(shape, dtype=np.float32) * u.DN / u.s

    # add dq array
    wfi_image.dq = np.zeros(shape, dtype=np.uint32)

    # construct ImageModel
    mod = ImageModel(wfi_image)

    filt = mod.meta.instrument["optical_element"]
    detector = mod.meta["instrument"]["detector"].replace("WFI", "SCA")

    # input parameters for PSF model:
    webbpsf_config = dict(
        filt=filt,
        detector=detector,
        oversample=12,
        fov_pixels=15,
    )

    dir_path = tempfile.gettempdir()
    filename_prefix = f"psf_model_{webbpsf_config['filt']}"
    file_path = os.path.join(dir_path, filename_prefix)

    # compute gridded PSF model:
    psf_model, centroids = create_gridded_psf_model(
        file_path,
        webbpsf_config["filt"],
        webbpsf_config["detector"],
        oversample=webbpsf_config["oversample"],
        fov_pixels=webbpsf_config["fov_pixels"],
        overwrite=True,
        logging_level="ERROR",
    )

    return mod, webbpsf_config, psf_model


def add_sources(image_model, psf_model, x_true, y_true, amp_true, background=10):
    shape = image_model.data.shape

    x_true, y_true, amp_true = (
        np.atleast_1d(val) for val in [x_true, y_true, amp_true]
    )

    fit_models = []
    for amp, x, y in zip(amp_true, x_true, y_true):
        model = deepcopy(psf_model)
        model.parameters = [amp, x, y]
        fit_models.append(model)

    photometry = PSFPhotometry(psf_model, (15, 15))
    photometry.fit_results = dict(local_bkg=np.ones(len(x_true)) * 0)
    photometry._fit_models = fit_models

    image = photometry.make_model_image(shape, (19, 19))

    image += rng.normal(background, 1, size=shape)
    image_model.data = image * np.ones_like(image_model.data)
    image_model.err = background * np.ones_like(image_model.err)


class TestPSFFitting:
    def setup_method(self):
        self.image_model, self.webbpsf_config, self.psf_model = setup_inputs(
            shape=image_model_shape,
        )

    @pytest.mark.webbpsf
    @pytest.mark.parametrize(
        "dx, dy, true_amp",
        zip(
            rng.uniform(-1, 1, n_trials),
            rng.uniform(-1, 1, n_trials),
            np.geomspace(1_000, 100_000, n_trials),
        ),
    )
    def test_psf_fit(self, dx, dy, true_amp):
        # generate an ImageModel
        image_model = deepcopy(self.image_model)
        init_data_stddev = np.std(image_model.data.value)

        # add synthetic sources to the ImageModel:
        true_x = image_model_shape[0] / 2 + dx
        true_y = image_model_shape[1] / 2 + dy
        add_sources(image_model, self.psf_model, true_x, true_y, true_amp)

        if self.webbpsf_config["fov_pixels"] % 2 == 0:
            fit_shape = (
                self.webbpsf_config["fov_pixels"] + 1,
                self.webbpsf_config["fov_pixels"] + 1,
            )
        else:
            fit_shape = (
                self.webbpsf_config["fov_pixels"],
                self.webbpsf_config["fov_pixels"],
            )

        # fit the PSF to the ImageModel:
        results_table, photometry = fit_psf_to_image_model(
            image_model=image_model,
            photometry_cls=PSFPhotometry,
            psf_model=self.psf_model,
            x_init=true_x,
            y_init=true_y,
            fit_shape=fit_shape,
        )

        # difference between input and output, normalized by the
        # uncertainty. Has units of sigma:
        delta_x = np.abs(true_x - results_table["x_fit"]) / results_table["x_err"]
        delta_y = np.abs(true_y - results_table["y_fit"]) / results_table["y_err"]

        sigma_threshold = 3.5
        assert np.all(delta_x < sigma_threshold)
        assert np.all(delta_y < sigma_threshold)

        # now check that the uncertainties aren't way too large, which could cause
        # the above test to pass even when the fits are bad. Use overly-simple
        # approximation that astrometric uncertainty be proportional to the
        # PSF's FWHM / SNR:
        approx_snr = true_amp / init_data_stddev
        approx_fwhm = 1
        approx_centroid_err = approx_fwhm / approx_snr

        # centroid err heuristic above is an underestimate, so we scale it up:
        scale_factor_approx = 100

        assert np.all(
            results_table["x_err"] < scale_factor_approx * approx_centroid_err
        )
        assert np.all(
            results_table["y_err"] < scale_factor_approx * approx_centroid_err
        )



"""

Unit tests for saturation flagging

"""

import numpy as np
import pytest
from astropy import units as u
from roman_datamodels import maker_utils
from roman_datamodels.datamodels import ScienceRawModel
from roman_datamodels.dqflags import group, pixel

from romancal.saturation import SaturationStep
from romancal.saturation.saturation import flag_saturation


def test_basic_saturation_flagging(setup_wfi_datamodels):
    """Check that the saturation flag is set when a pixel value is above the
    threshold given by the reference file."""

    # Create inputs, and data and saturation models
    ngroups = 5
    nrows = 20
    ncols = 20
    satvalue = 60000
    ramp, satmap = setup_wfi_datamodels(ngroups, nrows, ncols)

    # Add ramp values up to the saturation limit
    ramp.data[0, 5, 5] = 0 * ramp.data.unit
    ramp.data[1, 5, 5] = 20000 * ramp.data.unit
    ramp.data[2, 5, 5] = 40000 * ramp.data.unit
    ramp.data[3, 5, 5] = 60000 * ramp.data.unit  # Signal reaches saturation limit
    ramp.data[4, 5, 5] = 62000 * ramp.data.unit

    # Set saturation value in the saturation model
    satmap.data[5, 5] = satvalue * satmap.data.unit

    # Run the pipeline
    output = flag_saturation(ramp, satmap)

    # Make sure that groups with signal > saturation limit get flagged
    satindex = np.argmax(output.data.value[:, 5, 5] == satvalue)
    assert np.all(output.groupdq[satindex:, 5, 5] == group.SATURATED)


def test_read_pattern_saturation_flagging(setup_wfi_datamodels):
    """Check that the saturation threshold varies depending on how the reads
    are allocated into resultants."""

    # Create inputs, and data and saturation models
    ngroups = 5
    nrows = 20
    ncols = 20
    satvalue = 60000
    ramp, satmap = setup_wfi_datamodels(ngroups, nrows, ncols)

    # Add ramp values up to the saturation limit
    ramp.data[0, 5, 5] = 0 * ramp.data.unit
    ramp.data[1, 5, 5] = 20000 * ramp.data.unit
    ramp.data[2, 5, 5] = 40000 * ramp.data.unit
    ramp.data[3, 5, 5] = 60000 * ramp.data.unit  # Signal reaches saturation limit
    ramp.data[4, 5, 5] = 62000 * ramp.data.unit

    # set read_pattern to have many reads in the third resultant, so that
    # its mean exposure time is much smaller than its last read time
    # (in this case, the ratio is 13 / 20).
    # This means that the effective saturation for the third resultant
    # is 60000 * 13 / 20 = 39000 and the third resultant should be marked
    # saturated.
    ramp.meta.exposure.read_pattern = [
        [1],
        [2],
        [3, 4, 5, 6, 7, 8, 9, 10],
        [11],
        [12],
        [13],
    ]

    # Set saturation value in the saturation model
    satmap.data[5, 5] = satvalue * satmap.data.unit

    # Run the pipeline
    output = flag_saturation(ramp, satmap)

    # Make sure that groups after the third get flagged
    assert np.all(output.groupdq[2:, 5, 5] == group.SATURATED)


def test_ad_floor_flagging(setup_wfi_datamodels):
    """Check that the ad_floor flag is set when a pixel value is zero or
    negative."""

    # Create inputs, and data and saturation models
    ngroups = 5
    nrows = 20
    ncols = 20
    satvalue = 60000

    ramp, satmap = setup_wfi_datamodels(ngroups, nrows, ncols)

    # Add ramp values up to the saturation limit
    ramp.data[0, 5, 5] = 0 * ramp.data.unit  # Signal at bottom rail - low saturation
    ramp.data[1, 5, 5] = 0 * ramp.data.unit  # Signal at bottom rail - low saturation
    ramp.data[2, 5, 5] = 20 * ramp.data.unit
    ramp.data[3, 5, 5] = 40 * ramp.data.unit
    ramp.data[4, 5, 5] = 60 * ramp.data.unit

    # frames that should be flagged as saturation (low)
    satindxs = [0, 1]

    # Set saturation value in the saturation model
    satmap.data[5, 5] = satvalue * satmap.data.unit

    # Run the pipeline
    output = flag_saturation(ramp, satmap)

    # Check if the right frames are flagged as saturated
    assert np.all(output.groupdq[satindxs, 5, 5] == group.DO_NOT_USE | group.AD_FLOOR)


def test_ad_floor_and_saturation_flagging(setup_wfi_datamodels):
    """
    Check that the ad_floor flag is set when a pixel value is zero or
    negative and the saturation flag when the pixel is above the saturation threshold.
    """

    # Create inputs, and data and saturation models
    ngroups = 5
    nrows = 20
    ncols = 20
    satvalue = 60000

    ramp, satmap = setup_wfi_datamodels(ngroups, nrows, ncols)

    # Add ramp values up to the saturation limit
    ramp.data[0, 5, 5] = 0 * ramp.data.unit  # Signal at bottom rail - low saturation
    ramp.data[1, 5, 5] = 0 * ramp.data.unit  # Signal at bottom rail - low saturation
    ramp.data[2, 5, 5] = 20 * ramp.data.unit
    ramp.data[3, 5, 5] = 40 * ramp.data.unit
    ramp.data[4, 5, 5] = 61000 * ramp.data.unit  # Signal above the saturation threshold

    # frames that should be flagged as ad_floor
    floorindxs = [0, 1]
    # frames that should be flagged as saturation
    satindxs = [4]

    # Set saturation value in the saturation model
    satmap.data[5, 5] = satvalue * satmap.data.unit

    # Run the pipeline
    output = flag_saturation(ramp, satmap)

    # Check if the right frames are flagged as ad_floor
    assert np.all(output.groupdq[floorindxs, 5, 5] == group.DO_NOT_USE | group.AD_FLOOR)
    # Check if the right frames are flagged as saturated
    assert np.all(output.groupdq[satindxs, 5, 5] == group.SATURATED)


def test_signal_fluctuation_flagging(setup_wfi_datamodels):
    """Check that once a pixel is flagged as saturated in a group, all
    subsequent groups should also be flagged as saturated, even if the
    signal value drops back below saturation."""

    # Create inputs, and data and saturation models
    ngroups = 5
    nrows = 20
    ncols = 20
    satvalue = 60000

    ramp, satmap = setup_wfi_datamodels(ngroups, nrows, ncols)

    # Add ramp values up to the saturation limit
    ramp.data[0, 5, 5] = 10 * ramp.data.unit
    ramp.data[1, 5, 5] = 20000 * ramp.data.unit
    ramp.data[2, 5, 5] = 40000 * ramp.data.unit
    ramp.data[3, 5, 5] = 60000 * ramp.data.unit  # Signal reaches saturation limit
    ramp.data[4, 5, 5] = 40000 * ramp.data.unit  # Signal drops below saturation limit

    # Set saturation value in the saturation model
    satmap.data[5, 5] = satvalue * satmap.data.unit

    # Run the pipeline
    output = flag_saturation(ramp, satmap)

    # Make sure that all groups after first saturated group are flagged
    satindex = np.argmax(output.data.value[:, 5, 5] == satvalue)
    assert np.all(output.groupdq[satindex:, 5, 5] == group.SATURATED)


def test_all_groups_saturated(setup_wfi_datamodels):
    """Check case where all groups are saturated."""

    # Create inputs, and data and saturation models
    ngroups = 5
    nrows = 20
    ncols = 20
    satvalue = 60000

    ramp, satmap = setup_wfi_datamodels(ngroups, nrows, ncols)

    # Add ramp values at or above saturation limit
    ramp.data[0, 5, 5] = 60000 * ramp.data.unit
    ramp.data[1, 5, 5] = 62000 * ramp.data.unit
    ramp.data[2, 5, 5] = 62000 * ramp.data.unit
    ramp.data[3, 5, 5] = 60000 * ramp.data.unit
    ramp.data[4, 5, 5] = 62000 * ramp.data.unit

    # Set saturation value in the saturation model
    satmap.data[5, 5] = satvalue * satmap.data.unit

    # Run the pipeline
    output = flag_saturation(ramp, satmap)

    # Make sure all groups are flagged
    assert np.all(output.groupdq[:, 5, 5] == group.SATURATED)


def test_dq_propagation(setup_wfi_datamodels):
    """Check PIXELDQ propagation."""

    # Create inputs, data, and saturation maps
    ngroups = 5
    nrows = 20
    ncols = 20
    dqval1 = 5
    dqval2 = 10

    ramp, satmap = setup_wfi_datamodels(ngroups, nrows, ncols)

    # Add DQ values to the data and reference file
    ramp.pixeldq[5, 5] = dqval1
    satmap.dq[5, 5] = dqval2

    # Run the pipeline
    output = flag_saturation(ramp, satmap)

    # Make sure DQ values from data and reference file are added in the output
    assert output.pixeldq[5, 5] == dqval1 + dqval2


def test_no_sat_check(setup_wfi_datamodels):
    """Check that pixels flagged with NO_SAT_CHECK in the reference file get
    added to the DQ mask and are not flagged as saturated."""

    # Create inputs, and data and saturation models
    ngroups = 5
    nrows = 20
    ncols = 20
    satvalue = 60000

    ramp, satmap = setup_wfi_datamodels(ngroups, nrows, ncols)

    # Add ramp values up to the saturation limit
    ramp.data[0, 5, 5] = 10 * ramp.data.unit
    ramp.data[1, 5, 5] = 20000 * ramp.data.unit
    ramp.data[2, 5, 5] = 40000 * ramp.data.unit
    ramp.data[3, 5, 5] = 60000 * ramp.data.unit
    ramp.data[4, 5, 5] = 62000 * ramp.data.unit  # Signal reaches saturation limit

    # Set saturation value in the saturation model & DQ value for NO_SAT_CHECK
    satmap.data[5, 5] = satvalue * satmap.data.unit
    satmap.dq[5, 5] = pixel.NO_SAT_CHECK

    # Also set an existing DQ flag in input science data
    ramp.pixeldq[5, 5] = pixel.DO_NOT_USE

    # Run the pipeline
    output = flag_saturation(ramp, satmap)

    # Make sure output GROUPDQ does not get flagged as saturated
    # Make sure PIXELDQ is set to NO_SAT_CHECK and original flag
    assert np.all(output.groupdq[:, 5, 5] != group.SATURATED)
    # Test that saturation bit is NOT set
    assert np.all(
        output.groupdq[:, 5, 5] & (1 << group.SATURATED.bit_length() - 1) == 0
    )
    assert output.pixeldq[5, 5] == (pixel.NO_SAT_CHECK + pixel.DO_NOT_USE)


def test_nans_in_mask(setup_wfi_datamodels):
    """Check that pixels in the reference files that have value NaN are not
    flagged as saturated in the data and that in the PIXELDQ array the
    pixel is set to NO_SAT_CHECK."""

    # Create inputs, and data and saturation models
    ngroups = 5
    nrows = 20
    ncols = 20

    ramp, satmap = setup_wfi_datamodels(ngroups, nrows, ncols)

    # Add ramp values up to the saturation limit
    ramp.data[0, 5, 5] = 10 * ramp.data.unit
    ramp.data[1, 5, 5] = 20000 * ramp.data.unit
    ramp.data[2, 5, 5] = 40000 * ramp.data.unit
    ramp.data[3, 5, 5] = 60000 * ramp.data.unit
    ramp.data[4, 5, 5] = 62000 * ramp.data.unit

    # Set saturation value for pixel to NaN
    satmap.data[5, 5] = np.nan * satmap.data.unit

    # Run the pipeline
    output = flag_saturation(ramp, satmap)

    # Check that output GROUPDQ is not flagged as saturated
    assert np.all(output.groupdq[:, 5, 5] != group.SATURATED)
    # Check that output PIXELDQ is set to NO_SAT_CHECK
    assert output.pixeldq[5, 5] == pixel.NO_SAT_CHECK


def test_saturation_getbestref(setup_wfi_datamodels):
    """Check that when CRDS returns N/A for the reference file the
    step is skipped"""

    # Set test size
    shape = (2, 20, 20)

    # Create test science raw model
    wfi_sci_raw = maker_utils.mk_level1_science_raw(shape=shape)
    wfi_sci_raw.meta.instrument.name = "WFI"
    wfi_sci_raw.meta.instrument.detector = "WFI01"
    wfi_sci_raw.meta.instrument.optical_element = "F158"
    wfi_sci_raw.meta["guidestar"]["gw_window_xstart"] = 1012
    wfi_sci_raw.meta["guidestar"]["gw_window_xsize"] = 16
    wfi_sci_raw.meta.exposure.type = "WFI_IMAGE"
    wfi_sci_raw.data = u.Quantity(
        np.ones(shape, dtype=np.uint16), u.DN, dtype=np.uint16
    )
    wfi_sci_raw_model = ScienceRawModel(wfi_sci_raw, dq=True)

    # Run the pipeline
    result = SaturationStep.call(wfi_sci_raw_model, override_saturation="N/A")
    assert result.meta.cal_step.saturation == "SKIPPED"


@pytest.fixture(scope="function")
def setup_wfi_datamodels():
    """Set up fake WFI data to test."""

    def _models(ngroups, nrows, ncols):
        # Create ramp data
        ramp_model = maker_utils.mk_ramp(shape=(ngroups, nrows, ncols))

        # Create saturation reference data
        saturation_model = maker_utils.mk_saturation(shape=(nrows, ncols))

        return ramp_model, saturation_model

    return _models



import contextlib
import copy
import logging
import os
import os.path as op
import re
from collections import OrderedDict
from collections.abc import Iterable, Sequence
from pathlib import Path

import numpy as np
from roman_datamodels import datamodels as rdm

from romancal.lib.basic_utils import is_association

from ..associations import AssociationNotValidError, load_asn

__all__ = [
    "ModelContainer",
]

_ONE_MB = 1 << 20
RECOGNIZED_MEMBER_FIELDS = [
    "tweakreg_catalog",
]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ModelContainer(Sequence):
    """
    A container for holding DataModels.

    This functions like a list for holding DataModel objects.  It can be
    iterated through like a list and the datamodels within the container can be
    addressed by index. Additionally, the datamodels can be grouped by exposure.

    Parameters
    ----------
    init : path to ASN file, list of either datamodels or path to ASDF files, or `None`
        If `None`, then an empty `ModelContainer` instance is initialized, to which
        datamodels can later be added via the ``insert()``, ``append()``,
        or ``extend()`` method.

    iscopy : bool
        Presume this model is a copy. Members will not be closed
        when the model is closed/garbage-collected.

    memmap : bool
        Open ASDF file binary data using memmap (default: False)

    return_open : bool
        (optional) See notes below on usage.

    save_open : bool
        (optional) See notes below on usage.

    Examples
    --------
    To load a list of ASDF files into a `ModelContainer`:

    .. code-block:: python

        container = ModelContainer(
            [
                "/path/to/file1.asdf",
                "/path/to/file2.asdf",
                ...,
                "/path/to/fileN.asdf"
            ]
        )

    To load a list of open Roman DataModels into a `ModelContainer`:

    .. code-block:: python

        import roman_datamodels.datamodels as rdm
        data_list = [
                "/path/to/file1.asdf",
                "/path/to/file2.asdf",
                ...,
                "/path/to/fileN.asdf"
            ]
        datamodels_list = [rdm.open(x) for x in data_list]
        container = ModelContainer(datamodels_list)

    To load an ASN file into a `ModelContainer`:

    .. code-block:: python

        asn_file = "/path/to/asn_file.json"
        container = ModelContainer(asn_file)


    In any of the cases above, the content of each file in a `ModelContainer` can
    be accessed by iterating over its elements. For example, to print out the filename
    of each file, we can run:

    .. code-block:: python

        for model in container:
            print(model.meta.filename)


    Additionally, `ModelContainer` can be used with context manager:

    .. code-block:: python

        with ModelContainer(asn_file) as asn:
            # do stuff


    Notes
    -----
    The optional parameters ``save_open`` and ``return_open`` can be
    provided to control how the `DataModel` are used by the
    :py:class:`ModelContainer`. If ``save_open`` is set to `False`, each input
    `DataModel` instance in ``init`` will be written out to disk and
    closed, then only the filename for the `DataModel` will be used to
    initialize the :py:class:`ModelContainer` object.
    Subsequent access of each member will then open the `DataModel` file to
    work with it. If ``return_open`` is also `False`, then the `DataModel`
    will be closed when access to the `DataModel` is completed. The use of
    these parameters can minimize the amount of memory used by this object
    during processing.

    .. warning:: Input files will be updated in-place with new ``meta`` attribute
        values when ASN table's members contain additional attributes.

    """

    def __init__(
        self,
        init=None,
        asn_exptypes=None,
        asn_n_members=None,
        iscopy=False,
        memmap=False,
        # always return an open datamodel
        return_open=True,
        save_open=True,
    ):
        self._models = []
        self._iscopy = iscopy
        self._memmap = memmap
        self._return_open = return_open
        self._save_open = save_open

        self.asn_exptypes = asn_exptypes
        self.asn_n_members = asn_n_members
        self.asn_table = {}
        self.asn_table_name = None
        self.asn_pool_name = None
        self.filepaths = None

        try:
            init = Path(init)
        except TypeError:
            if init is None:
                # don't populate container
                pass
            elif isinstance(init, Sequence):
                # only append list items to self._models if all items are either
                # not-null strings (i.e. path to an ASDF file) or instances of DataModel
                is_all_string = all(isinstance(x, str) and len(x) for x in init)
                is_all_roman_datamodels = all(
                    isinstance(x, rdm.DataModel) for x in init
                )
                is_all_path = all(isinstance(x, Path) for x in init)

                if len(init) and (is_all_string or is_all_roman_datamodels):
                    self._models.extend(init)
                elif len(init) and is_all_path:
                    # parse Path object to string
                    self._models.extend([str(x) for x in init])
                else:
                    raise TypeError(
                        "Input must be an ASN file or a list of either strings "
                        "(full path to ASDF files) or Roman datamodels."
                    )
                if is_all_string or is_all_path:
                    self.filepaths = [op.basename(m) for m in self._models]
                else:
                    self.filepaths = getattr(init, "filepaths", None)
        else:
            if is_association(init):
                self.from_asn(init)
            elif isinstance(init, Path) and init.name != "":
                try:
                    init_from_asn = self.read_asn(init)
                    self.from_asn(init_from_asn, asn_file_path=init)
                except Exception as e:
                    raise TypeError(
                        "Input must be an ASN file or a list of either strings "
                        "(full path to ASDF files) or Roman datamodels."
                    ) from e
            else:
                raise TypeError(
                    "Input must be an ASN file or a list of either strings "
                    "(full path to ASDF files) or Roman datamodels."
                )

    def __len__(self):
        return len(self._models)

    def __getitem__(self, index):
        if isinstance(index, slice):
            start = index.start
            stop = index.stop
            step = index.step
            m = self._models[start:stop:step]
            m = [
                (
                    rdm.open(item, memmap=self._memmap)
                    if (not isinstance(item, rdm.DataModel) and self._return_open)
                    else item
                )
                for item in m
            ]
        else:
            m = self._models[index]
            if not isinstance(m, rdm.DataModel) and self._return_open:
                m = rdm.open(m, memmap=self._memmap)
        return m

    def __setitem__(self, index, model):
        if isinstance(model, rdm.DataModel):
            self._models[index] = model
        else:
            raise ValueError("Only datamodels can be used.")

    def __delitem__(self, index):
        del self._models[index]

    def __iter__(self):
        for model in self._models:
            if not isinstance(model, rdm.DataModel) and self._return_open:
                model = rdm.open(model, memmap=self._memmap)
            yield model

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # clean up
        for model in self._models:
            if isinstance(model, rdm.DataModel):
                model.close()
        # exceptions will be propagated out of the context
        return False

    def insert(self, index, model):
        if isinstance(model, rdm.DataModel):
            self._models.insert(index, model)
        else:
            raise ValueError("Only datamodels can be used.")

    def append(self, model):
        if isinstance(model, rdm.DataModel):
            self._models.append(model)
        else:
            raise ValueError("Only datamodels can be used.")

    def extend(self, input_object):
        if not isinstance(input_object, (Iterable, rdm.DataModel)) or isinstance(
            input_object, str
        ):
            raise ValueError("Not a valid input object.")
        elif all(isinstance(x, rdm.DataModel) for x in input_object):
            self._models.extend(input_object)
        else:
            raise ValueError("Not a valid input object.")

    def pop(self, index=-1):
        self._models.pop(index)

    def copy(self, memo=None):
        """
        Returns a deep copy of the models in this model container.
        """
        return copy.deepcopy(self, memo=memo)

    def close(self):
        """Close all datamodels."""
        if not self._iscopy:
            for model in self._models:
                if isinstance(model, rdm.DataModel):
                    model.close()

    @staticmethod
    def read_asn(filepath):
        """
        Load ASDF files from a Roman association file.

        Parameters
        ----------
        filepath : str
            The path to an association file.
        """
        filepath = op.abspath(op.expanduser(op.expandvars(filepath)))
        try:
            with open(filepath) as asn_file:
                asn_data = load_asn(asn_file)
        except AssociationNotValidError as e:
            raise OSError("Cannot read ASN file.") from e
        return asn_data

    def from_asn(self, asn_data, asn_file_path=None):
        """
        Load ASDF files from a Roman association file.

        Parameters
        ----------
        asn_data : `~roman_datamodels.associations.Association`
            Association dictionary.

        asn_file_path : str
            Filepath of the association, if known.
        """
        # match the asn_exptypes to the exptype in the association and retain
        # only those file that match, as a list, if asn_exptypes is set to none
        # grab all the files
        if self.asn_exptypes:
            infiles = []
            logger.debug(
                f"Filtering datasets based on allowed exptypes {self.asn_exptypes}:"
            )
            for member in asn_data["products"][0]["members"]:
                if any(
                    x
                    for x in self.asn_exptypes
                    if re.match(member["exptype"], x, re.IGNORECASE)
                ):
                    infiles.append(member)
                    logger.debug(f'Files accepted for processing {member["expname"]}:')
        else:
            infiles = list(asn_data["products"][0]["members"])

        asn_dir = op.dirname(asn_file_path) if asn_file_path else ""
        # Only handle the specified number of members.
        sublist = infiles[: self.asn_n_members] if self.asn_n_members else infiles
        self.filepaths = []
        try:
            for member in sublist:
                filepath = op.join(asn_dir, member["expname"])
                self.filepaths.append(op.basename(filepath))
                update_model = any(attr in member for attr in RECOGNIZED_MEMBER_FIELDS)
                if update_model or self._save_open:
                    m = rdm.open(filepath, memmap=self._memmap)
                    m.meta["asn"] = {"exptype": member["exptype"]}
                    for attr, val in member.items():
                        if attr in RECOGNIZED_MEMBER_FIELDS:
                            if attr == "tweakreg_catalog":
                                val = op.join(asn_dir, val) if val.strip() else None
                            m.meta[attr] = val

                    if not self._save_open:
                        m.save(filepath)
                        m.close()
                else:
                    m = filepath

                self._models.append(m)

        except OSError:
            self.close()
            raise

        # Pull the whole association table into asn_table
        self.merge_tree(self.asn_table, asn_data)

        if asn_file_path is not None:
            self.asn_table_name = op.basename(asn_file_path)
            self.asn_pool_name = asn_data["asn_pool"]
            for model in self:
                with contextlib.suppress(AttributeError):
                    model.meta.asn["table_name"] = self.asn_table_name
                    model.meta.asn["pool_name"] = self.asn_pool_name

    def save(self, path=None, dir_path=None, save_model_func=None, **kwargs):
        """
        Write out models in container to ASDF.

        Parameters
        ----------
        path : str or func or None
            - If None, the `meta.filename` is used for each model.
            - If a string, the string is used as a root and an index is
              appended.
            - If a function, the function takes the two arguments:
              the value of model.meta.filename and the
              `idx` index, returning constructed file name.

        dir_path : str
            Directory to write out files.  Defaults to current working dir.
            If directory does not exist, it creates it.  Filenames are pulled
            from `.meta.filename` of each datamodel in the container.

        save_model_func: func or None
            Alternate function to save each model instead of
            the models `save` method. Takes one argument, the model,
            and keyword argument `idx` for an index.

        Note
        ----
        Additional parameters provided via `**kwargs` are passed on to
        `roman_datamodels.datamodels.DataModel.to_asdf`

        Returns
        -------
        output_paths: [str[, ...]]
            List of output file paths of where the models were saved.
        """
        output_paths = []
        if path is None:

            def path(filename, idx=None):
                return filename

        elif not callable(path):
            path = make_file_with_index

        # use current path if dir_path is not provided
        dir_path = dir_path if dir_path is not None else os.getcwd()
        # output filename suffix
        output_suffix = kwargs.pop("output_suffix", None)
        for idx, model in enumerate(self._models):
            if len(self) <= 1:
                idx = None
            if save_model_func is None:
                filename = model.meta.filename
                output_path, output_filename = op.split(path(filename, idx=idx))

                # use dir_path when provided
                output_path = output_path if dir_path is None else dir_path

                # handle optional modifications to filename
                base, ext = op.splitext(output_filename)
                if output_suffix is not None:
                    # add suffix to filename
                    base = "".join([base, output_suffix])
                output_filename = "".join([base, ext])

                # create final destination (path + filename)
                save_path = op.join(output_path, output_filename)

                if ext == ".asdf":
                    output_paths.append(save_path)
                    model.to_asdf(save_path, **kwargs)
                else:
                    raise ValueError(f"Unknown filetype {ext}")
            else:
                output_paths.append(save_model_func(model, idx=idx))

        return output_paths

    @property
    def models_grouped(self):
        """
        Returns a list of a list of datamodels grouped by exposure.
        Assign an ID grouping by exposure.

        Data from different detectors of the same exposure will have the
        same group id, which allows grouping by exposure.  The following
        metadata is used for grouping:

        meta.observation.program
        meta.observation.observation
        meta.observation.visit
        meta.observation.visit_file_group
        meta.observation.visit_file_sequence
        meta.observation.visit_file_activity
        meta.observation.exposure
        """
        unique_exposure_parameters = [
            "program",
            "observation",
            "visit",
            "visit_file_group",
            "visit_file_sequence",
            "visit_file_activity",
            "exposure",
        ]

        group_dict = OrderedDict()
        for i, model in enumerate(self._models):
            model = model if isinstance(model, rdm.DataModel) else rdm.open(model)

            if not self._save_open:
                model = rdm.open(model, memmap=self._memmap)

            params = [
                str(getattr(model.meta.observation, param))
                for param in unique_exposure_parameters
            ]
            try:
                group_id = "roman" + "_".join(
                    ["".join(params[:3]), "".join(params[3:6]), params[6]]
                )
                model.meta["group_id"] = group_id
            except TypeError:
                model.meta["group_id"] = f"exposure{i + 1:04d}"

            group_id = model.meta.group_id
            if not self._save_open and not self._return_open:
                model.close()
                model = self._models[i]

            if group_id in group_dict:
                group_dict[group_id].append(model)
            else:
                group_dict[group_id] = [model]

        return group_dict.values()

    def merge_tree(self, a, b):
        """
        Merge elements from tree ``b`` into tree ``a``.
        """

        def recurse(a, b):
            if isinstance(b, dict):
                if not isinstance(a, dict):
                    return copy.deepcopy(b)
                for key, val in b.items():
                    a[key] = recurse(a.get(key), val)
                return a
            return copy.deepcopy(b)

        recurse(a, b)
        return a

    @property
    def crds_observatory(self):
        """
        Get the CRDS observatory for this container.  Used when selecting
        step/pipeline parameter files when the container is a pipeline input.

        Returns
        -------
        str
        """
        return "roman"

    def get_crds_parameters(self):
        """
        Get parameters used by CRDS to select references for this model.

        Returns
        -------
        dict
        """
        crds_header = {}
        if len(self._models):
            model = self._models[0]
            model = model if isinstance(model, rdm.DataModel) else rdm.open(model)
            crds_header |= model.get_crds_parameters()

        return crds_header

    def set_buffer(self, buffer_size, overlap=None):
        """Set buffer size for scrolling section-by-section access.

        Parameters
        ----------
        buffer_size : float, None
            Define size of buffer in MB for each section.
            If `None`, a default buffer size of 1MB will be used.

        overlap : int, optional
            Define the number of rows of overlaps between sections.
            If `None`, no overlap will be used.
        """
        self.overlap = 0 if overlap is None else overlap
        self.grow = 0

        with rdm.open(self._models[0]) as model:
            imrows, imcols = model.data.shape
            data_item_size = model.data.itemsize
            data_item_type = model.data.dtype
            del model

        min_buffer_size = imcols * data_item_size

        self.buffer_size = (
            min_buffer_size if buffer_size is None else (buffer_size * _ONE_MB)
        )

        section_nrows = min(imrows, int(self.buffer_size // min_buffer_size))

        if section_nrows == 0:
            self.buffer_size = min_buffer_size
            logger.warning(
                "WARNING: Buffer size is too small to hold a single row."
                f"Increasing buffer size to {self.buffer_size / _ONE_MB}MB"
            )
            section_nrows = 1

        nbr = section_nrows - self.overlap
        nsec = (imrows - self.overlap) // nbr
        if (imrows - self.overlap) % nbr > 0:
            nsec += 1

        self.n_sections = nsec
        self.nbr = nbr
        self.section_nrows = section_nrows
        self.imrows = imrows
        self.imcols = imcols
        self.imtype = data_item_type

    def get_sections(self):
        """Iterator to return the sections from all members of the container."""

        for k in range(self.n_sections):
            e1 = k * self.nbr
            e2 = e1 + self.section_nrows

            if k == self.n_sections - 1:  # last section
                e2 = min(e2, self.imrows)
                e1 = min(e1, e2 - self.overlap - 1)

            data_list = np.empty(
                (len(self._models), e2 - e1, self.imcols), dtype=self.imtype
            )
            wht_list = np.empty(
                (len(self._models), e2 - e1, self.imcols), dtype=self.imtype
            )
            for i, model in enumerate(self._models):
                model = rdm.open(model, memmap=self._memmap)

                data_list[i, :, :] = model.data[e1:e2].copy()
                wht_list[i, :, :] = model.weight[e1:e2].copy()
                del model

            yield (data_list, wht_list, (e1, e2))


def make_file_with_index(file_path, idx):
    """Append an index to a filename

    Parameters
    ----------
    file_path: str
        The file to append the index to.
    idx: int
        An index to append


    Returns
    -------
    file_path: str
        Path with index appended
    """
    # Decompose path
    path_head, path_tail = op.split(file_path)
    base, ext = op.splitext(path_tail)
    if idx is not None:
        base = base + str(idx)
    return op.join(path_head, base + ext)



from pathlib import Path

import pytest

from romancal.datamodels import filetype

DATA_DIRECTORY = Path(__file__).parent / "data"


def test_filetype():
    file_1 = filetype.check(DATA_DIRECTORY / "empty.json")
    file_2 = filetype.check(DATA_DIRECTORY / "example_schema.json")
    with open(DATA_DIRECTORY / "fake.json") as file_h:
        file_3 = filetype.check(file_h)
    file_4 = filetype.check(DATA_DIRECTORY / "empty.asdf")
    file_5 = filetype.check(DATA_DIRECTORY / "pluto.asdf")
    with open(DATA_DIRECTORY / "pluto.asdf", "rb") as file_h:
        file_6 = filetype.check(file_h)
    file_7 = filetype.check(DATA_DIRECTORY / "fake.asdf")
    with open(DATA_DIRECTORY / "fake.json") as file_h:
        file_8 = filetype.check(file_h)
    file_9 = filetype.check(str(DATA_DIRECTORY / "pluto.asdf"))

    assert file_1 == "asn"
    assert file_2 == "asn"
    assert file_3 == "asn"
    assert file_4 == "asdf"
    assert file_5 == "asdf"
    assert file_6 == "asdf"
    assert file_7 == "asdf"
    assert file_8 == "asn"
    assert file_9 == "asdf"

    with pytest.raises(ValueError):
        filetype.check(DATA_DIRECTORY / "empty.txt")

    with pytest.raises(ValueError):
        filetype.check(2)

    with pytest.raises(ValueError):
        filetype.check("test")



"""
Module for the source catalog step.
"""

import numpy as np
from astropy.table import Table
from roman_datamodels import datamodels, maker_utils
from roman_datamodels.datamodels import ImageModel, MosaicModel

from romancal.source_catalog.background import RomanBackground
from romancal.source_catalog.detection import convolve_data, make_segmentation_image
from romancal.source_catalog.reference_data import ReferenceData
from romancal.source_catalog.source_catalog import RomanSourceCatalog
from romancal.stpipe import RomanStep

__all__ = ["SourceCatalogStep"]


class SourceCatalogStep(RomanStep):
    """
    Create a catalog of sources including photometry and basic shape
    measurements.

    Parameters
    -----------
    input : str, `ImageModel`, or `MosaicModel`
        Path to an ASDF file, or an `ImageModel` or `MosaicModel`.
    """

    class_alias = "source_catalog"
    reference_file_types = []

    spec = """
        bkg_boxsize = integer(default=1000)   # background mesh box size in pixels
        kernel_fwhm = float(default=2.0)      # Gaussian kernel FWHM in pixels
        snr_threshold = float(default=3.0)    # per-pixel SNR threshold above the bkg
        npixels = integer(default=25)         # min number of pixels in source
        deblend = boolean(default=False)      # deblend sources?
        aperture_ee1 = integer(default=30)    # aperture encircled energy 1
        aperture_ee2 = integer(default=50)    # aperture encircled energy 2
        aperture_ee3 = integer(default=70)    # aperture encircled energy 3
        ci1_star_threshold = float(default=2.0)  # CI 1 star threshold
        ci2_star_threshold = float(default=1.8)  # CI 2 star threshold
        suffix = string(default='cat')        # Default suffix for output files
    """

    def process(self, input_model):
        with datamodels.open(input_model) as model:
            if not isinstance(model, (ImageModel, MosaicModel)):
                raise ValueError(
                    "The input model must be an ImageModel or MosaicModel."
                )

            if isinstance(model, ImageModel):
                cat_model = datamodels.SourceCatalogModel
            else:
                cat_model = datamodels.MosaicSourceCatalogModel
            source_catalog_model = maker_utils.mk_datamodel(cat_model)

            for key in source_catalog_model.meta.keys():
                try:
                    if key == "optical_element":
                        value = model.meta.instrument[key]
                    else:
                        value = model.meta[key]
                    source_catalog_model.meta[key] = value
                except KeyError:
                    pass

            aperture_ee = (self.aperture_ee1, self.aperture_ee2, self.aperture_ee3)
            refdata = ReferenceData(model, aperture_ee)
            aperture_params = refdata.aperture_params

            mask = np.isnan(model.data)
            coverage_mask = np.isnan(model.err)
            bkg = RomanBackground(
                model.data,
                box_size=self.bkg_boxsize,
                mask=mask,
                coverage_mask=coverage_mask,
            )
            model.data -= bkg.background

            convolved_data = convolve_data(
                model.data, kernel_fwhm=self.kernel_fwhm, mask=coverage_mask
            )

            segment_img = make_segmentation_image(
                convolved_data,
                snr_threshold=self.snr_threshold,
                npixels=self.npixels,
                bkg_rms=bkg.background_rms,
                deblend=self.deblend,
                mask=coverage_mask,
            )

            if segment_img is None:  # no sources found
                source_catalog_model.source_catalog = Table()
                return source_catalog_model

            ci_star_thresholds = (self.ci1_star_threshold, self.ci2_star_threshold)
            catobj = RomanSourceCatalog(
                model,
                segment_img,
                convolved_data,
                aperture_params,
                ci_star_thresholds,
                self.kernel_fwhm,
            )

            # put the resulting catalog in the model
            source_catalog_model.source_catalog = catobj.catalog

            # add back background to data so input model is unchanged
            # (in case of interactive use)
            model.data += bkg.background

            if self.save_results:
                # NOTE: the source_catalog_model is automatically saved
                #       if save_results = True

                # save the segmentation map
                segmentation_model = maker_utils.mk_datamodel(
                    datamodels.MosaicSegmentationMapModel
                )
                segmentation_model.data = segment_img.data.astype(np.uint32)
                self.save_model(segmentation_model, suffix="segm")

        return source_catalog_model



import glob
import io
import os

S3_BUCKET_NAME = "test-s3-data"


class MockS3Client:
    def __init__(self, s3_test_data_path):
        self.s3_test_data_path = s3_test_data_path

    def get_object(self, bucket_name, key):
        assert self.object_exists(bucket_name, key)

        with open(self._get_path(key), "rb") as f:
            return io.BytesIO(f.read())

    def object_exists(self, bucket_name, key):
        if bucket_name != S3_BUCKET_NAME:
            return False

        return os.path.isfile(self._get_path(key))

    def prefix_exists(self, bucket_name, key_prefix):
        return any(self.iterate_keys(bucket_name, key_prefix))

    def iterate_keys(self, bucket_name, key_prefix):
        if bucket_name != S3_BUCKET_NAME:
            return

        for k in self._list_keys():
            if k.startswith(key_prefix):
                yield k

    def _get_path(self, key):
        return os.path.join(self.s3_test_data_path, key)

    def _list_keys(self):
        paths = glob.glob(self.s3_test_data_path + "/**", recursive=True)
        paths = [p for p in paths if os.path.isfile(p)]
        return [p.replace(self.s3_test_data_path, "") for p in paths]


# Check strings based on words using length precision
def word_precision_check(str1, str2, length=5):
    """Check to strings word-by-word based for word length

    The strings are checked word for word, but only for the first
    `length` characters

    Parameters
    ----------
    str1, str2: str
        The strings to compare

    length: int
        The number of characters in each word to check.

    Returns
    -------
    match: boolean
        True if the strings match

    Raises
    ------
    AssertionError
        Raised if the number of word differ or the beginning components of each word differ.
    """
    words1 = str1.split()
    words2 = str2.split()
    if len(words1) != len(words2):
        raise AssertionError(
            f"str1 has different number of words {len(words1)} than str2 {len(words2)}"
        )
    for w1, w2 in zip(words1, words2):
        if w1[:length] != w2[:length]:
            raise AssertionError(f"str1 word {w1[:length]} != str2 {w2[:length]}")
    return True



#!/usr/bin/env python

import argparse

import pytest


def main():
    """Import all packages/modules in jwst to test they are importable

    This script is used in a Travis build that doesn't install any test
    dependencies from the package. This is to verify that all the modules
    are at least importable with only the dependencies listed in
    install_requires in setup.py.

    This is to prevent adding code to the runtime codebase that have
    dependencies that are only pulled in via the test dependencies.  So for
    example, accidentally adding a runtime dependency on pytest.
    """

    parser = argparse.ArgumentParser(
        description="Check if install_requires is up-to-date"
    )
    parser.parse_args()

    pytest.main(
        [
            "romancal/tests/test_import.py",
            "-c",
            "romancal/tests/empty_config/pytest.ini",
        ]
    )


if __name__ == "__main__":
    main()
